@@ -1,32 +1,33 @@
 #![allow(clippy::uninlined_format_args)]
 
-use rsconf::{LinkType, Target};
+use fish_build_helper::{fish_build_dir, workspace_root};
+use rsconf::Target;
 use std::env;
 use std::error::Error;
 use std::path::{Path, PathBuf};
 
+fn canonicalize<P: AsRef<Path>>(path: P) -> PathBuf {
+    std::fs::canonicalize(path).unwrap()
+}
+
 fn main() {
     setup_paths();
 
     // Add our default to enable tools that don't go through CMake, like "cargo test" and the
     // language server.
 
-    // FISH_BUILD_DIR is set by CMake, if we are using it.
-    // OUT_DIR is set by Cargo when the build script is running (not compiling)
-    let default_build_dir = env::var("OUT_DIR").unwrap();
-    let build_dir = option_env!("FISH_BUILD_DIR").unwrap_or(&default_build_dir);
-    let build_dir = std::fs::canonicalize(build_dir).unwrap();
-    let build_dir = build_dir.to_str().unwrap();
-    rsconf::set_env_value("FISH_BUILD_DIR", build_dir);
+    rsconf::set_env_value(
+        "FISH_BUILD_DIR",
+        // If set by CMake, this might include symlinks. Since we want to compare this to the
+        // dir fish is executed in we need to canonicalize it.
+        canonicalize(fish_build_dir()).to_str().unwrap(),
+    );
+
     // We need to canonicalize (i.e. realpath) the manifest dir because we want to be able to
     // compare it directly as a string at runtime.
     rsconf::set_env_value(
         "CARGO_MANIFEST_DIR",
-        std::fs::canonicalize(env!("CARGO_MANIFEST_DIR"))
-            .unwrap()
-            .as_path()
-            .to_str()
-            .unwrap(),
+        canonicalize(workspace_root()).to_str().unwrap(),
     );
 
     // Some build info
@@ -41,23 +42,19 @@ fn main() {
 
     std::env::set_var("FISH_BUILD_VERSION", version);
 
-    #[cfg(feature = "installable")]
-    #[cfg(not(clippy))]
-    {
-        let cman = std::fs::canonicalize(env!("CARGO_MANIFEST_DIR")).unwrap();
-        let targetman = cman.as_path().join("target").join("man");
-        build_man(&targetman);
-    }
-    rsconf::rebuild_if_path_changed("src/libc.c");
-    cc::Build::new()
-        .file("src/libc.c")
-        .include(build_dir)
-        .compile("flibc.a");
+    // These are necessary if built with embedded functions,
+    // but only in release builds (because rust-embed in debug builds reads from the filesystem).
+    #[cfg(feature = "embed-data")]
+    #[cfg(any(windows, not(debug_assertions)))]
+    rsconf::rebuild_if_path_changed("share");
 
-    let mut build = cc::Build::new();
-    // Add to the default library search path
-    build.flag_if_supported("-L/usr/local/lib/");
-    rsconf::add_library_search_path("/usr/local/lib");
+    #[cfg(feature = "gettext-extract")]
+    rsconf::rebuild_if_env_changed("FISH_GETTEXT_EXTRACTION_FILE");
+
+    rsconf::rebuild_if_path_changed("src/libc.c");
+    cc::Build::new().file("src/libc.c").compile("flibc.a");
+
+    let build = cc::Build::new();
     let mut target = Target::new_from(build).unwrap();
     // Keep verbose mode on until we've ironed out rust build script stuff
     target.set_verbose(true);
@@ -88,8 +85,9 @@ fn detect_cfgs(target: &mut Target) {
             "",
             &(|_: &Target| Ok(false)) as &dyn Fn(&Target) -> Result<bool, Box<dyn Error>>,
         ),
+        ("apple", &detect_apple),
         ("bsd", &detect_bsd),
-        ("gettext", &have_gettext),
+        ("cygwin", &detect_cygwin),
         ("small_main_stack", &has_small_stack),
         // See if libc supports the thread-safe localeconv_l(3) alternative to localeconv(3).
         ("localeconv_l", &|target| {
@@ -123,6 +121,15 @@ fn detect_cfgs(target: &mut Target) {
     }
 }
 
+fn detect_apple(_: &Target) -> Result<bool, Box<dyn Error>> {
+    Ok(cfg!(any(target_os = "ios", target_os = "macos")))
+}
+
+fn detect_cygwin(_: &Target) -> Result<bool, Box<dyn Error>> {
+    // Cygwin target is usually cross-compiled.
+    Ok(std::env::var("CARGO_CFG_TARGET_OS").unwrap() == "cygwin")
+}
+
 /// Detect if we're being compiled for a BSD-derived OS, allowing targeting code conditionally with
 /// `#[cfg(bsd)]`.
 ///
@@ -147,51 +154,6 @@ fn detect_bsd(_: &Target) -> Result<bool, Box<dyn Error>> {
     Ok(is_bsd)
 }
 
-/// Detect libintl/gettext and its needed symbols to enable internationalization/localization
-/// support.
-fn have_gettext(target: &Target) -> Result<bool, Box<dyn Error>> {
-    // The following script correctly detects and links against gettext, but so long as we are using
-    // C++ and generate a static library linked into the C++ binary via CMake, we need to account
-    // for the CMake option WITH_GETTEXT being explicitly disabled.
-    rsconf::rebuild_if_env_changed("CMAKE_WITH_GETTEXT");
-    if let Some(with_gettext) = std::env::var_os("CMAKE_WITH_GETTEXT") {
-        if with_gettext.eq_ignore_ascii_case("0") {
-            return Ok(false);
-        }
-    }
-
-    // In order for fish to correctly operate, we need some way of notifying libintl to invalidate
-    // its localizations when the locale environment variables are modified. Without the libintl
-    // symbol _nl_msg_cat_cntr, we cannot use gettext even if we find it.
-    let mut libraries = Vec::new();
-    let mut found = 0;
-    let symbols = ["gettext", "_nl_msg_cat_cntr"];
-    for symbol in &symbols {
-        // Historically, libintl was required in order to use gettext() and co, but that
-        // functionality was subsumed by some versions of libc.
-        if target.has_symbol(symbol) {
-            // No need to link anything special for this symbol
-            found += 1;
-            continue;
-        }
-        for library in ["intl", "gettextlib"] {
-            if target.has_symbol_in(symbol, &[library]) {
-                libraries.push(library);
-                found += 1;
-                continue;
-            }
-        }
-    }
-    match found {
-        0 => Ok(false),
-        1 => Err(format!("gettext found but cannot be used without {}", symbols[1]).into()),
-        _ => {
-            rsconf::link_libraries(&libraries, LinkType::Default);
-            Ok(true)
-        }
-    }
-}
-
 /// Rust sets the stack size of newly created threads to a sane value, but is at at the mercy of the
 /// OS when it comes to the size of the main stack. Some platforms we support default to a tiny
 /// 0.5 MiB main stack, which is insufficient for fish's MAX_EVAL_DEPTH/MAX_STACK_DEPTH values.
@@ -199,14 +161,14 @@ fn have_gettext(target: &Target) -> Result<bool, Box<dyn Error>> {
 /// 0.5 MiB is small enough that we'd have to drastically reduce MAX_STACK_DEPTH to less than 10, so
 /// we instead use a workaround to increase the main thread size.
 fn has_small_stack(_: &Target) -> Result<bool, Box<dyn Error>> {
-    #[cfg(not(any(target_os = "macos", target_os = "netbsd")))]
+    #[cfg(not(any(target_os = "ios", target_os = "macos", target_os = "netbsd")))]
     return Ok(false);
 
     // NetBSD 10 also needs this but can't find pthread_get_stacksize_np.
     #[cfg(target_os = "netbsd")]
     return Ok(true);
 
-    #[cfg(target_os = "macos")]
+    #[cfg(any(target_os = "ios", target_os = "macos"))]
     {
         use core::ffi;
 
@@ -227,6 +189,9 @@ fn has_small_stack(_: &Target) -> Result<bool, Box<dyn Error>> {
 }
 
 fn setup_paths() {
+    #[cfg(windows)]
+    use unix_path::{Path, PathBuf};
+
     fn get_path(name: &str, default: &str, onvar: &Path) -> PathBuf {
         let mut var = PathBuf::from(env::var(name).unwrap_or(default.to_string()));
         if var.is_relative() {
@@ -235,53 +200,44 @@ fn setup_paths() {
         var
     }
 
-    let (prefix_from_home, prefix) = if let Ok(pre) = env::var("PREFIX") {
-        (false, PathBuf::from(pre))
-    } else {
-        (true, PathBuf::from(".local/"))
-    };
-
-    // If someone gives us a $PREFIX, we need it to be absolute.
-    // Otherwise we would try to get it from $HOME and that won't really work.
-    if !prefix_from_home && prefix.is_relative() {
-        panic!("Can't have relative prefix");
-    }
-
+    let prefix = PathBuf::from(env::var("PREFIX").unwrap_or("/usr/local".to_string()));
     rsconf::rebuild_if_env_changed("PREFIX");
     rsconf::set_env_value("PREFIX", prefix.to_str().unwrap());
 
     let datadir = get_path("DATADIR", "share/", &prefix);
-    rsconf::set_env_value("DATADIR", datadir.to_str().unwrap());
-    rsconf::rebuild_if_env_changed("DATADIR");
-
-    let datadir_subdir = if prefix_from_home {
-        "fish/install"
-    } else {
-        "fish"
-    };
-    rsconf::set_env_value("DATADIR_SUBDIR", datadir_subdir);
-
-    let bindir = get_path("BINDIR", "bin/", &prefix);
-    rsconf::set_env_value("BINDIR", bindir.to_str().unwrap());
-    rsconf::rebuild_if_env_changed("BINDIR");
 
     let sysconfdir = get_path(
         "SYSCONFDIR",
-        // If we get our prefix from $HOME, we should use the system's /etc/
-        // ~/.local/share/etc/ makes no sense
-        if prefix_from_home { "/etc/" } else { "etc/" },
+        // Embedded builds use "/etc," not "./share/etc".
+        if cfg!(feature = "embed-data") {
+            "/etc/"
+        } else {
+            "etc/"
+        },
         &datadir,
     );
     rsconf::set_env_value("SYSCONFDIR", sysconfdir.to_str().unwrap());
     rsconf::rebuild_if_env_changed("SYSCONFDIR");
 
-    let localedir = get_path("LOCALEDIR", "locale/", &datadir);
-    rsconf::set_env_value("LOCALEDIR", localedir.to_str().unwrap());
-    rsconf::rebuild_if_env_changed("LOCALEDIR");
+    #[cfg(not(feature = "embed-data"))]
+    {
+        rsconf::set_env_value("DATADIR", datadir.to_str().unwrap());
+        rsconf::rebuild_if_env_changed("DATADIR");
 
-    let docdir = get_path("DOCDIR", "doc/fish", &datadir);
-    rsconf::set_env_value("DOCDIR", docdir.to_str().unwrap());
-    rsconf::rebuild_if_env_changed("DOCDIR");
+        let bindir = get_path("BINDIR", "bin/", &prefix);
+        rsconf::set_env_value("BINDIR", bindir.to_str().unwrap());
+        rsconf::rebuild_if_env_changed("BINDIR");
+
+        let localedir = get_path("LOCALEDIR", "locale/", &datadir);
+        let localedir = localedir.to_str().unwrap();
+        assert!(!localedir.is_empty(), "empty LOCALEDIR is not supported");
+        rsconf::set_env_value("LOCALEDIR", localedir);
+        rsconf::rebuild_if_env_changed("LOCALEDIR");
+
+        let docdir = get_path("DOCDIR", "doc/fish", &datadir);
+        rsconf::set_env_value("DOCDIR", docdir.to_str().unwrap());
+        rsconf::rebuild_if_env_changed("DOCDIR");
+    }
 }
 
 fn get_version(src_dir: &Path) -> String {
@@ -294,7 +250,7 @@ fn get_version(src_dir: &Path) -> String {
 
     let path = src_dir.join("version");
     if let Ok(strver) = read_to_string(path) {
-        return strver.to_string();
+        return strver;
     }
 
     let args = &["describe", "--always", "--dirty=-dirty"];
@@ -321,17 +277,37 @@ fn get_version(src_dir: &Path) -> String {
     // or because it refused (safe.directory applies to `git describe`!)
     // So we read the SHA ourselves.
     fn get_git_hash() -> Result<String, Box<dyn std::error::Error>> {
-        let gitdir = Path::new(env!("CARGO_MANIFEST_DIR")).join(".git");
+        let workspace_root = workspace_root();
+        let gitdir = workspace_root.join(".git");
+        let jjdir = workspace_root.join(".jj");
+        let commit_id = if gitdir.exists() {
+            // .git/HEAD contains ref: refs/heads/branch
+            let headpath = gitdir.join("HEAD");
+            let headstr = read_to_string(headpath)?;
+            let headref = headstr.split(' ').nth(1).unwrap().trim();
 
-        // .git/HEAD contains ref: refs/heads/branch
-        let headpath = gitdir.join("HEAD");
-        let headstr = read_to_string(headpath)?;
-        let headref = headstr.split(' ').collect::<Vec<_>>()[1].trim();
-
-        // .git/refs/heads/branch contains the SHA
-        let refpath = gitdir.join(headref);
-        // Shorten to 9 characters (what git describe does currently)
-        let refstr = &read_to_string(refpath)?[0..9];
+            // .git/refs/heads/branch contains the SHA
+            let refpath = gitdir.join(headref);
+            // Shorten to 9 characters (what git describe does currently)
+            read_to_string(refpath)?
+        } else if jjdir.exists() {
+            let output = Command::new("jj")
+                .args([
+                    "log",
+                    "--revisions",
+                    "@",
+                    "--no-graph",
+                    "--ignore-working-copy",
+                    "--template",
+                    "commit_id",
+                ])
+                .output()
+                .unwrap();
+            String::from_utf8_lossy(&output.stdout).to_string()
+        } else {
+            return Err("did not find either of .git or .jj".into());
+        };
+        let refstr = &commit_id[0..9];
         let refstr = refstr.trim();
 
         let version = env!("CARGO_PKG_VERSION").to_owned();
@@ -340,74 +316,3 @@ fn get_version(src_dir: &Path) -> String {
 
     get_git_hash().expect("Could not get a version. Either set $FISH_BUILD_VERSION or install git.")
 }
-
-#[cfg(feature = "installable")]
-// disable clippy because otherwise it would panic without sphinx
-#[cfg(not(clippy))]
-fn build_man(build_dir: &Path) {
-    use std::process::Command;
-    let mandir = build_dir;
-    let sec1dir = mandir.join("man1");
-    let docsrc_path = std::fs::canonicalize(env!("CARGO_MANIFEST_DIR"))
-        .unwrap()
-        .as_path()
-        .join("doc_src");
-    let docsrc = docsrc_path.to_str().unwrap();
-    let args = &[
-        "-j",
-        "auto",
-        "-q",
-        "-b",
-        "man",
-        "-c",
-        docsrc,
-        // doctree path - put this *above* the man1 dir to exclude it.
-        // this is ~6M
-        "-d",
-        mandir.to_str().unwrap(),
-        docsrc,
-        sec1dir.to_str().unwrap(),
-    ];
-    let _ = std::fs::create_dir_all(sec1dir.to_str().unwrap());
-
-    rsconf::rebuild_if_env_changed("FISH_BUILD_DOCS");
-    if env::var("FISH_BUILD_DOCS") == Ok("0".to_string()) {
-        println!("cargo:warning=Skipping man pages because $FISH_BUILD_DOCS is set to 0");
-        return;
-    }
-
-    // We run sphinx to build the man pages.
-    // Every error here is fatal so cargo doesn't cache the result
-    // - if we skipped the docs with sphinx not installed, installing it would not then build the docs.
-    // That means you need to explicitly set $FISH_BUILD_DOCS=0 (`FISH_BUILD_DOCS=0 cargo install --path .`),
-    // which is unfortunate - but the docs are pretty important because they're also used for --help.
-    match Command::new("sphinx-build").args(args).spawn() {
-        Err(x) if x.kind() == std::io::ErrorKind::NotFound => {
-            if env::var("FISH_BUILD_DOCS") == Ok("1".to_string()) {
-                panic!("Could not find sphinx-build to build man pages.\nInstall sphinx or disable building the docs by setting $FISH_BUILD_DOCS=0.");
-            }
-            println!("cargo:warning=Cannot find sphinx-build to build man pages.");
-            println!("cargo:warning=If you install it now you need to run `cargo clean` and rebuild, or set $FISH_BUILD_DOCS=1 explicitly.");
-        }
-        Err(x) => {
-            // Another error - permissions wrong etc
-            panic!("Error starting sphinx-build to build man pages: {:?}", x);
-        }
-        Ok(mut x) => match x.wait() {
-            Err(err) => {
-                panic!(
-                    "Error waiting for sphinx-build to build man pages: {:?}",
-                    err
-                );
-            }
-            Ok(out) => {
-                if out.success() {
-                    // Success!
-                    return;
-                } else {
-                    panic!("sphinx-build failed to build the man pages.");
-                }
-            }
-        },
-    }
-}
