@@ -20,6 +20,7 @@ fi
 for tool in \
     bundle \
     gh \
+    jq \
     ruby \
     timeout \
 ; do
@@ -34,7 +35,8 @@ fish_site=$repo_root/../fish-site
 
 for path in . "$fish_site"
 do
-    if ! git -C "$path" diff HEAD --quiet; then
+    if ! git -C "$path" diff HEAD --quiet ||
+        git ls-files --others --exclude-standard | grep .; then
         echo >&2 "$0: index and worktree must be clean"
         exit 1
     fi
@@ -45,7 +47,14 @@ if git tag | grep -qxF "$version"; then
     exit 1
 fi
 
-sed -n 1p CHANGELOG.rst | grep -q '^fish .*(released ???)$'
+integration_branch=$(
+    git for-each-ref --points-at=HEAD 'refs/heads/Integration_*' \
+        --format='%(refname:strip=2)'
+)
+[ -n "$integration_branch" ] ||
+    git merge-base --is-ancestor $remote/master HEAD
+
+sed -n 1p CHANGELOG.rst | grep -q '^fish .*(released .*)$'
 sed -n 2p CHANGELOG.rst | grep -q '^===*$'
 
 changelog_title="fish $version (released $(date +'%B %d, %Y'))"
@@ -70,10 +79,15 @@ git tag --annotate --message="Release $version" $version
 
 git push $remote $version
 
+TIMEOUT=
 gh() {
-    command gh --repo "$repository_owner/fish-shell" "$@"
+    command ${TIMEOUT:+timeout $TIMEOUT} \
+        gh --repo "$repository_owner/fish-shell" "$@"
 }
 
+gh workflow run release.yml --ref="$version" \
+    --raw-field="version=$version"
+
 run_id=
 while [ -z "$run_id" ] && sleep 5
 do
@@ -86,22 +100,27 @@ done
 # Update fishshell.com
 tag_oid=$(git rev-parse "$version")
 tmpdir=$(mktemp -d)
+# TODO This works on draft releases only if "gh" is configured to
+# have write access to the fish-shell repository. Unless we are fine
+# publishing the release at this point, we should at least fail if
+# "gh" doesn't have write access.
 while ! \
     gh release download "$version" --dir="$tmpdir" \
         --pattern="fish-$version.tar.xz"
 do
-    timeout 30 gh run watch "$run_id" ||:
+    TIMEOUT=30 gh run watch "$run_id" ||:
+    sleep 5
 done
 actual_tag_oid=$(git ls-remote "$remote" |
     awk '$2 == "refs/tags/'"$version"'" { print $1 }')
 [ "$tag_oid" = "$actual_tag_oid" ]
-tar -C "$tmpdir" xf fish-$version.tar.xz
-minor_version=${version%.*}
+( cd "$tmpdir" && tar xf fish-$version.tar.xz )
 CopyDocs() {
-    rm -rf "fish-site/site/docs/$1"
-    cp -r "$tmpdir/fish-$version/user_doc/html" "fish-site/site/docs/$1"
-    git -C fish-site add "site/docs/$1"
+    rm -rf "$fish_site/site/docs/$1"
+    cp -r "$tmpdir/fish-$version/user_doc/html" "$fish_site/site/docs/$1"
+    git -C $fish_site add "site/docs/$1"
 }
+minor_version=${version%.*}
 CopyDocs "$minor_version"
 latest_release=$(
     releases=$(git tag | grep '^[0-9]*\.[0-9]*\.[0-9]*.*' |
@@ -115,17 +134,43 @@ fi
 rm -rf "$tmpdir"
 (
     cd "$fish_site"
-    make new-release
+    make
     git add -u
+    ! git ls-files --others --exclude-standard | grep .
     git commit --message="$(printf %s "\
-        | Release $version
+        | Release $version (docs)
         |
         | Created by ../fish-shell/build_tools/release.sh
-    " | sed 's,^\s*| ,,')"
+    " | sed 's,^\s*| \?,,')"
 )
 
-# N.B. --exit-status doesn't fail reliably.
-gh run view "$run_id" --verbose --log-failed --exit-status
+# Approve macos-codesign
+# TODO what if current user can't approve?
+gh_pending_deployments() {
+    command gh api \
+        -H "Accept: application/vnd.github+json" \
+        -H "X-GitHub-Api-Version: 2022-11-28" \
+        "/repos/$repository_owner/fish-shell/actions/runs/$run_id/pending_deployments" \
+        "$@"
+}
+while {
+    environment_id=$(gh_pending_deployments | jq .[].environment.id)
+    [ -z "$environment_id" ]
+}
+do
+    sleep 5
+done
+echo '
+        {
+            "environment_ids": ['"$environment_id"'],
+            "state": "approved",
+            "comment": "Approved via ./build_tools/release.sh"
+        }
+    ' |
+gh_pending_deployments -XPOST --input=-
+
+# Await completion.
+gh run watch "$run_id"
 
 while {
     ! draft=$(gh release view "$version" --json=isDraft --jq=.isDraft) \
@@ -137,39 +182,47 @@ done
 
 (
     cd "$fish_site"
-    git push git@github.com:$repository_owner/fish-site
+    make new-release
+    git add -u
+    ! git ls-files --others --exclude-standard | grep .
+    git commit --message="$(printf %s "\
+        | Release $version (release list update)
+        |
+        | Created by ../fish-shell/build_tools/release.sh
+    " | sed 's,^\s*| \?,,')"
+    # This takes care to support remote names that are different from
+    # fish-shell remote name. Also, support detached HEAD state.
+    git push git@github.com:$repository_owner/fish-site HEAD:master
 )
 
-if git merge-base --is-ancestor $remote/master $version
-then
-    git push $remote $version:master
+if [ -n "$integration_branch" ]; then
+    git push $remote "$version^{commit}":refs/heads/$integration_branch
 else
-    # Probably on an integration branch.
-    # TODO Maybe push when that's safe (or move this to CI).
-    :
-fi
-
-if [ "$repository_owner" = fish-shell ]; then {
-    mail=$(mktemp)
-    cat >$mail <<EOF
-From: $(git var GIT_AUTHOR_IDENT | sed 's/ [0-9]* +[0-9]*$//')
-To: fish-users Mailing List <fish-users@lists.sourceforge.net>
-Subject: fish $version released
-
-See https://github.com/fish-shell/fish-shell/releases/tag/$version
-EOF
-    git send-email --suppress-cc=all $mail
-    rm $mail
-} fi
-
-changelog=$(cat - CHANGELOG.rst <<EOF
+    changelog=$(cat - CHANGELOG.rst <<EOF
 fish ?.?.? (released ???)
 =========================
 
 EOF
-)
-printf %s\\n "$changelog" >CHANGELOG.rst
-CommitVersion ${version}-snapshot "start new cycle"
+    )
+    printf %s\\n "$changelog" >CHANGELOG.rst
+    CommitVersion ${version}-snapshot "start new cycle"
+    git push $remote HEAD:master
+fi
+
+# TODO This can currently require a TTY for editing and password
+# prompts.
+if [ "$repository_owner" = fish-shell ]; then {
+    mail=$(mktemp)
+    cat >$mail <<EOF
+From: $(git var GIT_AUTHOR_IDENT | sed 's/ [0-9]* +[0-9]*$//')
+Subject: fish $version released
+
+See https://github.com/fish-shell/fish-shell/releases/tag/$version
+EOF
+    git send-email --suppress-cc=all --confirm=always $mail \
+        --to="fish-users Mailing List <fish-users@lists.sourceforge.net>"
+    rm $mail
+} fi
 
 exit
 
