@@ -124,7 +124,7 @@ fn join(list: &[&wstr], sep: &wstr) -> WString {
 }
 
 // Print abbreviations in a fish-script friendly way.
-fn abbr_show(streams: &mut IoStreams) -> Option<c_int> {
+fn abbr_show(streams: &mut IoStreams) -> BuiltinResult {
     let style = EscapeStringStyle::Script(Default::default());
 
     abbrs::with_abbrs(|abbrs| {
@@ -174,11 +174,11 @@ fn abbr_show(streams: &mut IoStreams) -> Option<c_int> {
         }
     });
 
-    return STATUS_CMD_OK;
+    return Ok(SUCCESS);
 }
 
 // Print the list of abbreviation names.
-fn abbr_list(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
+fn abbr_list(opts: &Options, streams: &mut IoStreams) -> BuiltinResult {
     const subcmd: &wstr = L!("--list");
     if !opts.args.is_empty() {
         streams.err.append(wgettext_fmt!(
@@ -187,7 +187,7 @@ fn abbr_list(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
             subcmd,
             &opts.args[0]
         ));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
     abbrs::with_abbrs(|abbrs| {
         for abbr in abbrs.list() {
@@ -197,11 +197,11 @@ fn abbr_list(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
         }
     });
 
-    return STATUS_CMD_OK;
+    return Ok(SUCCESS);
 }
 
 // Rename an abbreviation, deleting any existing one with the given name.
-fn abbr_rename(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
+fn abbr_rename(opts: &Options, streams: &mut IoStreams) -> BuiltinResult {
     const subcmd: &wstr = L!("--rename");
 
     if opts.args.len() != 2 {
@@ -210,7 +210,7 @@ fn abbr_rename(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
             CMD,
             subcmd
         ));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
     let old_name = &opts.args[0];
     let new_name = &opts.args[1];
@@ -220,7 +220,7 @@ fn abbr_rename(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
             CMD,
             subcmd
         ));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if contains_whitespace(new_name) {
@@ -230,9 +230,9 @@ fn abbr_rename(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
             subcmd,
             new_name.as_utfstr()
         ));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
-    abbrs::with_abbrs_mut(|abbrs| -> Option<c_int> {
+    abbrs::with_abbrs_mut(|abbrs| -> BuiltinResult {
         if !abbrs.has_name(old_name) {
             streams.err.append(wgettext_fmt!(
                 "%ls %ls: No abbreviation named %ls\n",
@@ -240,7 +240,7 @@ fn abbr_rename(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
                 subcmd,
                 old_name.as_utfstr()
             ));
-            return STATUS_CMD_ERROR;
+            return Err(STATUS_CMD_ERROR);
         }
         if abbrs.has_name(new_name) {
             streams.err.append(wgettext_fmt!(
@@ -250,10 +250,10 @@ fn abbr_rename(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
                 new_name.as_utfstr(),
                 old_name.as_utfstr()
             ));
-            return STATUS_INVALID_ARGS;
+            return Err(STATUS_INVALID_ARGS);
         }
         abbrs.rename(old_name, new_name);
-        STATUS_CMD_OK
+        Ok(SUCCESS)
     })
 }
 
@@ -262,20 +262,20 @@ fn contains_whitespace(val: &wstr) -> bool {
 }
 
 // Test if any args is an abbreviation.
-fn abbr_query(opts: &Options) -> Option<c_int> {
+fn abbr_query(opts: &Options) -> BuiltinResult {
     // Return success if any of our args matches an abbreviation.
     abbrs::with_abbrs(|abbrs| {
         for arg in opts.args.iter() {
             if abbrs.has_name(arg) {
-                return STATUS_CMD_OK;
+                return Ok(SUCCESS);
             }
         }
-        return STATUS_CMD_ERROR;
+        return Err(STATUS_CMD_ERROR);
     })
 }
 
 // Add a named abbreviation.
-fn abbr_add(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
+fn abbr_add(opts: &Options, streams: &mut IoStreams) -> BuiltinResult {
     const subcmd: &wstr = L!("--add");
 
     if opts.args.len() < 2 && opts.function.is_none() {
@@ -284,7 +284,7 @@ fn abbr_add(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
             CMD,
             subcmd
         ));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if opts.args.is_empty() || opts.args[0].is_empty() {
@@ -293,7 +293,7 @@ fn abbr_add(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
             CMD,
             subcmd
         ));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
     let name = &opts.args[0];
     if name.chars().any(|c| c.is_whitespace()) {
@@ -303,7 +303,7 @@ fn abbr_add(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
             subcmd,
             name.as_utfstr()
         ));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     let key: &wstr;
@@ -331,7 +331,7 @@ fn abbr_add(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
                     .err
                     .append(sprintf!("%ls: %*ls\n", CMD, offset, "^"));
             }
-            return STATUS_INVALID_ARGS;
+            return Err(STATUS_INVALID_ARGS);
         }
         let anchored = regex_make_anchored(regex_pattern);
         let re = Box::new(
@@ -352,7 +352,7 @@ fn abbr_add(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
         streams
             .err
             .append(wgettext_fmt!(BUILTIN_ERR_TOO_MANY_ARGUMENTS, L!("abbr")));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
     let replacement = if let Some(ref function) = opts.function {
         // Abbreviation function names disallow spaces.
@@ -363,7 +363,7 @@ fn abbr_add(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
                 CMD,
                 function.as_utfstr()
             ));
-            return STATUS_INVALID_ARGS;
+            return Err(STATUS_INVALID_ARGS);
         }
         function.clone()
     } else {
@@ -390,7 +390,7 @@ fn abbr_add(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
             "%ls: --command cannot be combined with --position command",
             CMD,
         ));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     // Note historically we have allowed overwriting existing abbreviations.
@@ -408,22 +408,22 @@ fn abbr_add(opts: &Options, streams: &mut IoStreams) -> Option<c_int> {
         })
     });
 
-    return STATUS_CMD_OK;
+    return Ok(SUCCESS);
 }
 
 // Erase the named abbreviations.
-fn abbr_erase(opts: &Options, parser: &Parser) -> Option<c_int> {
+fn abbr_erase(opts: &Options, parser: &Parser) -> BuiltinResult {
     if opts.args.is_empty() {
         // This has historically been a silent failure.
-        return STATUS_CMD_ERROR;
+        return Err(STATUS_CMD_ERROR);
     }
 
     // Erase each. If any is not found, return NotFound which is historical.
-    abbrs::with_abbrs_mut(|abbrs| -> Option<c_int> {
-        let mut result = STATUS_CMD_OK;
+    abbrs::with_abbrs_mut(|abbrs| -> BuiltinResult {
+        let mut result: BuiltinResult = Ok(SUCCESS);
         for arg in &opts.args {
             if !abbrs.erase(arg) {
-                result = Some(EnvStackSetResult::NotFound.into());
+                result = EnvStackSetResult::NotFound.into();
             }
             // Erase the old uvar - this makes `abbr -e` work.
             let esc_src = escape(arg);
@@ -432,7 +432,7 @@ fn abbr_erase(opts: &Options, parser: &Parser) -> Option<c_int> {
                 let ret = parser.vars().remove(&var_name, EnvMode::UNIVERSAL);
 
                 if ret == EnvStackSetResult::Ok {
-                    result = STATUS_CMD_OK
+                    result = Ok(SUCCESS)
                 };
             }
         }
@@ -440,7 +440,7 @@ fn abbr_erase(opts: &Options, parser: &Parser) -> Option<c_int> {
     })
 }
 
-pub fn abbr(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Option<c_int> {
+pub fn abbr(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> BuiltinResult {
     let mut argv_read = Vec::with_capacity(argv.len());
     argv_read.extend_from_slice(argv);
 
@@ -453,7 +453,7 @@ pub fn abbr(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
     // Note the leading '-' causes wgetopter to return arguments in order, instead of permuting
     // them. We need this behavior for compatibility with pre-builtin abbreviations where options
     // could be given literally, for example `abbr e emacs -nw`.
-    const short_options: &wstr = L!("-:ac:f:r:seqgUh");
+    const short_options: &wstr = L!("-ac:f:r:seqgUh");
 
     const longopts: &[WOption] = &[
         wopt(L!("add"), ArgType::NoArgument, 'a'),
@@ -504,7 +504,7 @@ pub fn abbr(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
                         "%ls: Cannot specify multiple positions\n",
                         CMD
                     ));
-                    return STATUS_INVALID_ARGS;
+                    return Err(STATUS_INVALID_ARGS);
                 }
                 if w.woptarg == Some(L!("command")) {
                     opts.position = Some(Position::Command);
@@ -519,7 +519,7 @@ pub fn abbr(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
                     streams
                         .err
                         .append(L!("Position must be one of: command, anywhere.\n"));
-                    return STATUS_INVALID_ARGS;
+                    return Err(STATUS_INVALID_ARGS);
                 }
             }
             'r' => {
@@ -528,7 +528,7 @@ pub fn abbr(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
                         "%ls: Cannot specify multiple regex patterns\n",
                         CMD
                     ));
-                    return STATUS_INVALID_ARGS;
+                    return Err(STATUS_INVALID_ARGS);
                 }
                 opts.regex_pattern = w.woptarg.map(ToOwned::to_owned);
             }
@@ -538,7 +538,7 @@ pub fn abbr(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
                         "%ls: Cannot specify multiple set-cursor options\n",
                         CMD
                     ));
-                    return STATUS_INVALID_ARGS;
+                    return Err(STATUS_INVALID_ARGS);
                 }
                 // The default set-cursor indicator is '%'.
                 let _ = opts
@@ -566,15 +566,19 @@ pub fn abbr(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
             }
             'h' => {
                 builtin_print_help(parser, streams, cmd);
-                return STATUS_CMD_OK;
+                return Ok(SUCCESS);
             }
             ':' => {
                 builtin_missing_argument(parser, streams, cmd, argv[w.wopt_index - 1], true);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
+            }
+            ';' => {
+                builtin_unexpected_argument(parser, streams, cmd, argv[w.wopt_index - 1], true);
+                return Err(STATUS_INVALID_ARGS);
             }
             '?' => {
                 builtin_unknown_option(parser, streams, cmd, argv[w.wopt_index - 1], false);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
             _ => {
                 panic!("unexpected retval from wgeopter.next()");
@@ -587,7 +591,7 @@ pub fn abbr(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
     }
 
     if !opts.validate(streams) {
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if opts.add {
