@@ -4,9 +4,11 @@ use crate::complete::{complete_add_wrapper, complete_remove_wrapper, CompletionR
 use crate::highlight::colorize;
 use crate::highlight::highlight_shell;
 use crate::nix::isatty;
+use crate::operation_context::OperationContext;
 use crate::parse_constants::ParseErrorList;
 use crate::parse_util::parse_util_detect_errors_in_argument_list;
 use crate::parse_util::{parse_util_detect_errors, parse_util_token_extent};
+use crate::proc::is_interactive_session;
 use crate::reader::{commandline_get_state, completion_apply_to_command_line};
 use crate::wcstringutil::string_suffixes_string;
 use crate::{
@@ -28,8 +30,8 @@ fn builtin_complete_add2(
     cmd: &wstr,
     cmd_is_path: bool,
     short_opt: &wstr,
-    gnu_opts: &[WString],
-    old_opts: &[WString],
+    gnu_opts: &[&wstr],
+    old_opts: &[&wstr],
     result_mode: CompletionMode,
     condition: &[WString],
     comp: &wstr,
@@ -50,7 +52,7 @@ fn builtin_complete_add2(
         );
     }
 
-    for gnu_opt in gnu_opts {
+    for &gnu_opt in gnu_opts {
         complete_add(
             cmd.to_owned(),
             cmd_is_path,
@@ -64,7 +66,7 @@ fn builtin_complete_add2(
         );
     }
 
-    for old_opt in old_opts {
+    for &old_opt in old_opts {
         complete_add(
             cmd.to_owned(),
             cmd_is_path,
@@ -98,8 +100,8 @@ fn builtin_complete_add(
     cmds: &[WString],
     paths: &[WString],
     short_opt: &wstr,
-    gnu_opt: &[WString],
-    old_opt: &[WString],
+    gnu_opt: &[&wstr],
+    old_opt: &[&wstr],
     result_mode: CompletionMode,
     condition: &[WString],
     comp: &wstr,
@@ -140,8 +142,8 @@ fn builtin_complete_remove_cmd(
     cmd: &WString,
     cmd_is_path: bool,
     short_opt: &wstr,
-    gnu_opt: &[WString],
-    old_opt: &[WString],
+    gnu_opt: &[&wstr],
+    old_opt: &[&wstr],
 ) {
     let mut removed = false;
     for s in short_opt.chars() {
@@ -184,8 +186,8 @@ fn builtin_complete_remove(
     cmds: &[WString],
     paths: &[WString],
     short_opt: &wstr,
-    gnu_opt: &[WString],
-    old_opt: &[WString],
+    gnu_opt: &[&wstr],
+    old_opt: &[&wstr],
 ) {
     for cmd in cmds {
         builtin_complete_remove_cmd(cmd, false /* not path */, short_opt, gnu_opt, old_opt);
@@ -215,17 +217,15 @@ fn builtin_complete_print(cmd: &wstr, streams: &mut IoStreams, parser: &Parser)
 const OPT_ESCAPE: char = '\x01';
 
 /// The complete builtin. Used for specifying programmable tab-completions. Calls the functions in
-/// complete.cpp for any heavy lifting.
-pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Option<c_int> {
+/// complete.rs for any heavy lifting.
+pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> BuiltinResult {
     let cmd = argv[0];
     let argc = argv.len();
     let mut result_mode = CompletionMode::default();
     let mut remove = false;
     let mut short_opt = WString::new();
-    // todo!("these whould be Vec<&wstr>");
     let mut gnu_opt = vec![];
     let mut old_opt = vec![];
-    let mut subcommand = vec![];
     let mut comp = WString::new();
     let mut desc = WString::new();
     let mut condition = vec![];
@@ -237,7 +237,7 @@ pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) ->
     let mut preserve_order = false;
     let mut unescape_output = true;
 
-    const short_options: &wstr = L!(":a:c:p:s:l:o:d:fFrxeuAn:C::w:hk");
+    const short_options: &wstr = L!("a:c:p:s:l:o:d:fFrxeuAn:C::w:hk");
     const long_options: &[WOption] = &[
         wopt(L!("exclusive"), ArgType::NoArgument, 'x'),
         wopt(L!("no-files"), ArgType::NoArgument, 'f'),
@@ -248,7 +248,6 @@ pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) ->
         wopt(L!("short-option"), ArgType::RequiredArgument, 's'),
         wopt(L!("long-option"), ArgType::RequiredArgument, 'l'),
         wopt(L!("old-option"), ArgType::RequiredArgument, 'o'),
-        wopt(L!("subcommand"), ArgType::RequiredArgument, 'S'),
         wopt(L!("description"), ArgType::RequiredArgument, 'd'),
         wopt(L!("arguments"), ArgType::RequiredArgument, 'a'),
         wopt(L!("erase"), ArgType::NoArgument, 'e'),
@@ -301,7 +300,7 @@ pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) ->
                         cmd,
                         w.woptarg.unwrap()
                     ));
-                    return STATUS_INVALID_ARGS;
+                    return Err(STATUS_INVALID_ARGS);
                 }
             }
             'd' => {
@@ -320,39 +319,30 @@ pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) ->
                     streams
                         .err
                         .append(wgettext_fmt!("%ls: -s requires a non-empty string\n", cmd,));
-                    return STATUS_INVALID_ARGS;
+                    return Err(STATUS_INVALID_ARGS);
                 }
             }
             'l' => {
                 let arg = w.woptarg.unwrap();
-                gnu_opt.push(arg.to_owned());
+                gnu_opt.push(arg);
                 if arg.is_empty() {
                     streams
                         .err
                         .append(wgettext_fmt!("%ls: -l requires a non-empty string\n", cmd,));
-                    return STATUS_INVALID_ARGS;
+                    return Err(STATUS_INVALID_ARGS);
                 }
             }
             'o' => {
                 let arg = w.woptarg.unwrap();
-                old_opt.push(arg.to_owned());
+                old_opt.push(arg);
                 if arg.is_empty() {
                     streams
                         .err
                         .append(wgettext_fmt!("%ls: -o requires a non-empty string\n", cmd,));
-                    return STATUS_INVALID_ARGS;
-                }
-            }
-            'S' => {
-                let arg = w.woptarg.unwrap();
-                subcommand.push(arg.to_owned());
-                if arg.is_empty() {
-                    streams
-                        .err
-                        .append(wgettext_fmt!("%ls: -S requires a non-empty string\n", cmd,));
-                    return STATUS_INVALID_ARGS;
+                    return Err(STATUS_INVALID_ARGS);
                 }
             }
+            'S' => {}
             'a' => {
                 comp = w.woptarg.unwrap().to_owned();
             }
@@ -374,15 +364,19 @@ pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) ->
             }
             'h' => {
                 builtin_print_help(parser, streams, cmd);
-                return STATUS_CMD_OK;
+                return Ok(SUCCESS);
             }
             ':' => {
                 builtin_missing_argument(parser, streams, cmd, argv[w.wopt_index - 1], true);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
+            }
+            ';' => {
+                builtin_unexpected_argument(parser, streams, cmd, argv[w.wopt_index - 1], true);
+                return Err(STATUS_INVALID_ARGS);
             }
             '?' => {
                 builtin_unknown_option(parser, streams, cmd, argv[w.wopt_index - 1], true);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
             _ => panic!("unexpected retval from WGetopter"),
         }
@@ -404,7 +398,7 @@ pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) ->
                 "'--exclusive' and '--force-files'"
             ));
         }
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if w.wopt_index != argc {
@@ -420,7 +414,7 @@ pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) ->
                 .err
                 .append(wgettext_fmt!(BUILTIN_ERR_TOO_MANY_ARGUMENTS, cmd));
             builtin_print_error_trailer(parser, streams.err, cmd);
-            return STATUS_INVALID_ARGS;
+            return Err(STATUS_INVALID_ARGS);
         }
     }
 
@@ -437,7 +431,7 @@ pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) ->
                 ));
                 streams.err.push('\n');
             }
-            return STATUS_CMD_ERROR;
+            return Err(STATUS_CMD_ERROR);
         }
     }
 
@@ -454,7 +448,7 @@ pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) ->
             ));
             streams.err.append(err_text);
             streams.err.push('\n');
-            return STATUS_CMD_ERROR;
+            return Err(STATUS_CMD_ERROR);
         }
     }
 
@@ -464,33 +458,28 @@ pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) ->
             None => {
                 // No argument given, try to use the current commandline.
                 let commandline_state = commandline_get_state(true);
-                if !commandline_state.initialized {
-                    // This corresponds to using 'complete -C' in non-interactive mode.
-                    // See #2361    .
-                    builtin_missing_argument(parser, streams, cmd, L!("-C"), true);
-                    return STATUS_INVALID_ARGS;
+                if !is_interactive_session() {
+                    streams.err.append(cmd);
+                    streams
+                        .err
+                        .append(L!(": Can not get commandline in non-interactive mode\n"));
+                    return Err(STATUS_CMD_ERROR);
                 }
                 commandline_state.text
             }
             Some(param) => param,
         };
 
-        let mut token = 0..0;
-        parse_util_token_extent(
-            &do_complete_param,
-            do_complete_param.len(),
-            &mut token,
-            None,
-        );
+        let (token, _) = parse_util_token_extent(&do_complete_param, do_complete_param.len());
 
         // Create a scoped transient command line, so that builtin_commandline will see our
         // argument, not the reader buffer.
-        parser
+        let saved_transient = parser
             .libdata_mut()
-            .transient_commandlines
-            .push(do_complete_param.clone());
+            .transient_commandline
+            .replace(do_complete_param.clone());
         let _remove_transient = ScopeGuard::new((), |()| {
-            parser.libdata_mut().transient_commandlines.pop();
+            parser.libdata_mut().transient_commandline = saved_transient;
         });
 
         // Prevent accidental recursion (see #6171).
@@ -513,6 +502,7 @@ pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) ->
                 let faux_cmdline = &do_complete_param[token.clone()];
                 let mut tmp_cursor = faux_cmdline.len();
                 let mut faux_cmdline_with_completion = completion_apply_to_command_line(
+                    &OperationContext::background_interruptible(parser.vars()),
                     &next.completion,
                     next.flags,
                     faux_cmdline,
@@ -615,5 +605,5 @@ pub fn complete(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) ->
         }
     }
 
-    STATUS_CMD_OK
+    Ok(SUCCESS)
 }
