@@ -40,7 +40,7 @@ impl Default for FunctionCmdOpts {
 
 // This command is atypical in using the "-" (RETURN_IN_ORDER) option for flag parsing.
 // This is needed due to the semantics of the -a/--argument-names flag.
-const SHORT_OPTIONS: &wstr = L!("-:a:d:e:hj:p:s:v:w:SV:");
+const SHORT_OPTIONS: &wstr = L!("-a:d:e:hj:p:s:v:w:SV:");
 #[rustfmt::skip]
 const LONG_OPTIONS: &[WOption] = &[
     wopt(L!("description"), ArgType::RequiredArgument, 'd'),
@@ -77,7 +77,7 @@ fn parse_cmd_opts(
     argv: &mut [&wstr],
     parser: &Parser,
     streams: &mut IoStreams,
-) -> Option<c_int> {
+) -> c_int {
     let cmd = L!("function");
     let print_hints = false;
     let mut handling_named_arguments = false;
@@ -142,9 +142,8 @@ fn parse_cmd_opts(
                 let woptarg = w.woptarg.unwrap();
                 let e: EventDescription;
                 if opt == 'j' && woptarg == "caller" {
-                    let libdata = parser.libdata();
-                    let caller_id = if libdata.is_subshell {
-                        libdata.caller_id
+                    let caller_id = if parser.scope().is_subshell {
+                        parser.scope().caller_id
                     } else {
                         0
                     };
@@ -220,6 +219,16 @@ fn parse_cmd_opts(
                 builtin_missing_argument(parser, streams, cmd, argv[w.wopt_index - 1], print_hints);
                 return STATUS_INVALID_ARGS;
             }
+            ';' => {
+                builtin_unexpected_argument(
+                    parser,
+                    streams,
+                    cmd,
+                    argv[w.wopt_index - 1],
+                    print_hints,
+                );
+                return STATUS_INVALID_ARGS;
+            }
             '?' => {
                 builtin_unknown_option(parser, streams, cmd, argv[w.wopt_index - 1], print_hints);
                 return STATUS_INVALID_ARGS;
@@ -239,7 +248,7 @@ fn validate_function_name(
     function_name: &mut WString,
     cmd: &wstr,
     streams: &mut IoStreams,
-) -> Option<c_int> {
+) -> c_int {
     if argv.len() < 2 {
         // This is currently impossible but let's be paranoid.
         streams
@@ -275,7 +284,7 @@ pub fn function(
     streams: &mut IoStreams,
     c_args: &mut [&wstr],
     func_node: NodeRef<BlockStatement>,
-) -> Option<c_int> {
+) -> c_int {
     // The wgetopt function expects 'function' as the first argument. Make a new vec with
     // that property. This is needed because this builtin has a different signature than the other
     // builtins.
@@ -355,7 +364,9 @@ pub fn function(
     let props = function::FunctionProperties {
         func_node,
         named_arguments: opts.named_arguments,
-        description: opts.description,
+        // Function descriptions are extracted from scripts in `share` via
+        // `build_tools/fish_xgettext.fish`.
+        description: LocalizableString::from_external_source(opts.description),
         inherit_vars: inherit_vars.into_boxed_slice(),
         shadow_scope: opts.shadow_scope,
         is_autoload: RelaxedAtomicBool::new(false),
