@@ -20,9 +20,13 @@ use crate::{
     wutil::wcstoi::wcstoi_partial,
 };
 
-const MISMATCHED_ARGS: &str = "%ls: given %d indexes but %d values\n";
-const UVAR_ERR: &str =
-    "%ls: successfully set universal '%ls'; but a global by that name shadows it\n";
+localizable_consts!(
+    MISMATCHED_ARGS
+    "%ls: given %d indexes but %d values\n"
+
+    UVAR_ERR
+    "%ls: successfully set universal '%ls'; but a global by that name shadows it\n"
+);
 
 #[derive(Debug, Clone)]
 struct Options {
@@ -96,7 +100,7 @@ impl Options {
         args: &mut [&wstr],
         parser: &Parser,
         streams: &mut IoStreams,
-    ) -> Result<(Options, usize), Option<c_int>> {
+    ) -> Result<Option<(Options, usize)>, ErrorCode> {
         /// Values used for long-only options.
         const PATH_ARG: char = 1 as char;
         const UNPATH_ARG: char = 2 as char;
@@ -104,7 +108,7 @@ impl Options {
         // Variables used for parsing the argument list. This command is atypical in using the "+"
         // (REQUIRE_ORDER) option for flag parsing. This is not typical of most fish commands. It means
         // we stop scanning for flags when the first non-flag argument is seen.
-        const SHORT_OPTS: &wstr = L!("+:LSUaefghlnpqux");
+        const SHORT_OPTS: &wstr = L!("+LSUaefghlnpqux");
         const LONG_OPTS: &[WOption] = &[
             wopt(L!("export"), NoArgument, 'x'),
             wopt(L!("global"), NoArgument, 'g'),
@@ -163,6 +167,16 @@ impl Options {
                     builtin_missing_argument(parser, streams, cmd, args[w.wopt_index - 1], false);
                     return Err(STATUS_INVALID_ARGS);
                 }
+                ';' => {
+                    builtin_unexpected_argument(
+                        parser,
+                        streams,
+                        cmd,
+                        args[w.wopt_index - 1],
+                        false,
+                    );
+                    return Err(STATUS_INVALID_ARGS);
+                }
                 '?' => {
                     // Specifically detect `set -o` because people might be bringing over bashisms.
                     let optind = w.wopt_index;
@@ -198,12 +212,12 @@ impl Options {
 
         if opts.print_help {
             builtin_print_help(parser, streams, cmd);
-            return Err(STATUS_CMD_OK);
+            return Ok(None);
         }
 
         Self::validate(&opts, cmd, args, optind, parser, streams)?;
 
-        Ok((opts, optind))
+        Ok(Some((opts, optind)))
     }
 
     fn validate(
@@ -213,7 +227,7 @@ impl Options {
         optind: usize,
         parser: &Parser,
         streams: &mut IoStreams,
-    ) -> Result<(), Option<c_int>> {
+    ) -> Result<(), ErrorCode> {
         // Can't query and erase or list.
         if opts.query && (opts.erase || opts.list) {
             streams.err.append(wgettext_fmt!(BUILTIN_ERR_COMBO, cmd));
@@ -530,7 +544,7 @@ fn erased_at_indexes(mut input: Vec<WString>, mut indexes: Vec<isize>) -> Vec<WS
 
 /// Print the names of all environment variables in the scope. It will include the values unless the
 /// `set --names` flag was used.
-fn list(opts: &Options, parser: &Parser, streams: &mut IoStreams) -> Option<c_int> {
+fn list(opts: &Options, parser: &Parser, streams: &mut IoStreams) -> BuiltinResult {
     let names_only = opts.list;
     let mut names = parser.vars().get_names(opts.scope());
     names.sort();
@@ -573,7 +587,7 @@ fn list(opts: &Options, parser: &Parser, streams: &mut IoStreams) -> Option<c_in
         streams.out.append(out);
     }
 
-    STATUS_CMD_OK
+    Ok(SUCCESS)
 }
 
 fn query(
@@ -582,20 +596,20 @@ fn query(
     parser: &Parser,
     streams: &mut IoStreams,
     args: &[&wstr],
-) -> Option<c_int> {
+) -> BuiltinResult {
     let mut retval = 0;
     let scope = opts.scope();
 
     // No variables given, this is an error.
     // 255 is the maximum return code we allow.
     if args.is_empty() {
-        return Some(255);
+        return Err(STATUS_NO_VARIABLES_GIVEN);
     }
 
     for arg in args {
         let Some(split) = split_var_and_indexes(arg, scope, parser.vars(), streams) else {
             builtin_print_error_trailer(parser, streams.err, cmd);
-            return STATUS_CMD_ERROR;
+            return Err(STATUS_CMD_ERROR);
         };
 
         if split.indexes.is_empty() {
@@ -614,7 +628,7 @@ fn query(
         }
     }
 
-    Some(retval)
+    BuiltinResult::from_dynamic(retval)
 }
 
 fn show_scope(var_name: &wstr, scope: EnvMode, streams: &mut IoStreams, vars: &dyn Environment) {
@@ -682,7 +696,7 @@ fn show_scope(var_name: &wstr, scope: EnvMode, streams: &mut IoStreams, vars: &d
 }
 
 /// Show mode. Show information about the named variable(s).
-fn show(cmd: &wstr, parser: &Parser, streams: &mut IoStreams, args: &[&wstr]) -> Option<c_int> {
+fn show(cmd: &wstr, parser: &Parser, streams: &mut IoStreams, args: &[&wstr]) -> BuiltinResult {
     let vars = parser.vars();
     if args.is_empty() {
         // show all vars
@@ -716,7 +730,7 @@ fn show(cmd: &wstr, parser: &Parser, streams: &mut IoStreams, args: &[&wstr]) ->
                     .err
                     .append(wgettext_fmt!(BUILTIN_ERR_VARNAME, cmd, arg));
                 builtin_print_error_trailer(parser, streams.err, cmd);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
 
             if arg.contains('[') {
@@ -725,7 +739,7 @@ fn show(cmd: &wstr, parser: &Parser, streams: &mut IoStreams, args: &[&wstr]) ->
                     cmd
                 ));
                 builtin_print_error_trailer(parser, streams.err, cmd);
-                return STATUS_CMD_ERROR;
+                return Err(STATUS_CMD_ERROR);
             }
 
             show_scope(arg, EnvMode::LOCAL, streams, vars);
@@ -745,7 +759,7 @@ fn show(cmd: &wstr, parser: &Parser, streams: &mut IoStreams, args: &[&wstr]) ->
         }
     }
 
-    STATUS_CMD_OK
+    Ok(SUCCESS)
 }
 
 fn erase(
@@ -754,8 +768,8 @@ fn erase(
     parser: &Parser,
     streams: &mut IoStreams,
     args: &[&wstr],
-) -> Option<c_int> {
-    let mut ret = STATUS_CMD_OK;
+) -> BuiltinResult {
+    let mut ret = Ok(SUCCESS);
     let scopes = opts.scope();
     // `set -e` is allowed to be called with multiple scopes.
     for bit in (0..).take_while(|bit| 1 << bit <= EnvMode::USER.bits()) {
@@ -766,7 +780,7 @@ fn erase(
         for arg in args {
             let Some(split) = split_var_and_indexes(arg, scope, parser.vars(), streams) else {
                 builtin_print_error_trailer(parser, streams.err, cmd);
-                return STATUS_CMD_ERROR;
+                return Err(STATUS_CMD_ERROR);
             };
 
             if !valid_var_name(split.varname) {
@@ -774,7 +788,7 @@ fn erase(
                     .err
                     .append(wgettext_fmt!(BUILTIN_ERR_VARNAME, cmd, split.varname));
                 builtin_print_error_trailer(parser, streams.err, cmd);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
             let retval;
             if split.indexes.is_empty() {
@@ -792,7 +806,7 @@ fn erase(
             } else {
                 // remove just the specified indexes of the var
                 let Some(var) = split.var else {
-                    return STATUS_CMD_ERROR;
+                    return Err(STATUS_CMD_ERROR);
                 };
                 let result = erased_at_indexes(var.as_list().to_owned(), split.indexes);
                 retval = env_set_reporting_errors(
@@ -809,23 +823,13 @@ fn erase(
             // Set $status to the last error value.
             // This is cheesy, but I don't expect this to be checked often.
             if retval != EnvStackSetResult::Ok {
-                ret = env_result_to_status(retval);
+                ret = retval.into();
             }
         }
     }
     ret
 }
 
-fn env_result_to_status(retval: EnvStackSetResult) -> Option<c_int> {
-    Some(match retval {
-        EnvStackSetResult::Ok => 0,
-        EnvStackSetResult::Perm => 1,
-        EnvStackSetResult::Scope => 2,
-        EnvStackSetResult::Invalid => 3,
-        EnvStackSetResult::NotFound => 4,
-    })
-}
-
 /// Return a list of new values for the variable `varname`, respecting the `opts`.
 /// This handles the simple case where there are no indexes.
 fn new_var_values(
@@ -897,16 +901,15 @@ fn set_internal(
     cmd: &wstr,
     opts: &Options,
     parser: &Parser,
-
     streams: &mut IoStreams,
     argv: &[&wstr],
-) -> Option<c_int> {
+) -> BuiltinResult {
     if argv.is_empty() {
         streams
             .err
             .append(wgettext_fmt!(BUILTIN_ERR_MIN_ARG_COUNT1, cmd, 1));
         builtin_print_error_trailer(parser, streams.err, cmd);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     let scope = opts.scope();
@@ -915,7 +918,7 @@ fn set_internal(
 
     let Some(split) = split_var_and_indexes(var_expr, scope, parser.vars(), streams) else {
         builtin_print_error_trailer(parser, streams.err, cmd);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     };
 
     // Is the variable valid?
@@ -932,7 +935,7 @@ fn set_internal(
             ));
         }
         builtin_print_error_trailer(parser, streams.err, cmd);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     // Setting with explicit indexes like `set foo[3] ...` has additional error handling.
@@ -944,7 +947,7 @@ fn set_internal(
                     .err
                     .append(wgettext_fmt!("%ls: array index out of bounds\n", cmd));
                 builtin_print_error_trailer(parser, streams.err, cmd);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
         }
         // Append and prepend are disallowed.
@@ -954,7 +957,7 @@ fn set_internal(
                 cmd
             ));
             builtin_print_error_trailer(parser, streams.err, cmd);
-            return STATUS_INVALID_ARGS;
+            return Err(STATUS_INVALID_ARGS);
         }
 
         // Argument count and index count must agree.
@@ -965,7 +968,7 @@ fn set_internal(
                 split.indexes.len(),
                 argv.len()
             ));
-            return STATUS_INVALID_ARGS;
+            return Err(STATUS_INVALID_ARGS);
         }
     }
 
@@ -984,15 +987,16 @@ fn set_internal(
     if retval == EnvStackSetResult::Ok {
         warn_if_uvar_shadows_global(cmd, opts, split.varname, streams, parser);
     }
-    env_result_to_status(retval)
+
+    retval.into()
 }
 
 /// The set builtin creates, updates, and erases (removes, deletes) variables.
-pub fn set(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
+pub fn set(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
     let cmd = args[0];
-    let (opts, optind) = match Options::parse(cmd, args, parser, streams) {
-        Ok(res) => res,
-        Err(err) => return err,
+    let (opts, optind) = match Options::parse(cmd, args, parser, streams)? {
+        Some((opts, optind)) => (opts, optind),
+        None => return Ok(SUCCESS),
     };
 
     let args = &args[optind..];
@@ -1011,8 +1015,10 @@ pub fn set(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Opti
         set_internal(cmd, &opts, parser, streams, args)
     };
 
-    if retval == STATUS_CMD_OK && opts.preserve_failure_exit_status {
-        return None;
+    if retval.is_ok() && opts.preserve_failure_exit_status {
+        return Ok(Success {
+            preserve_failure_exit_status: true,
+        });
     }
 
     return retval;
