@@ -1,7 +1,7 @@
 use super::prelude::*;
 use crate::builtins::*;
 use crate::common::{escape, get_by_sorted_name, str2wcstring, Named};
-use crate::io::{IoFd, OutputStream};
+use crate::io::OutputStream;
 use crate::parse_constants::UNKNOWN_BUILTIN_ERR_MSG;
 use crate::parse_util::parse_util_argument_is_help;
 use crate::parser::{Block, BlockType, LoopStatus};
@@ -9,97 +9,167 @@ use crate::proc::{no_exec, ProcStatus};
 use crate::reader::reader_read;
 use crate::wchar::L;
 use errno::errno;
-use libc::{STDERR_FILENO, STDIN_FILENO, STDOUT_FILENO};
+use libc::STDIN_FILENO;
 
 use std::fs::File;
 use std::io::{BufRead, BufReader, Read};
 use std::os::fd::FromRawFd;
-use std::sync::Arc;
 
-pub type BuiltinCmd = fn(&Parser, &mut IoStreams, &mut [&wstr]) -> Option<c_int>;
+pub type BuiltinCmd = fn(&Parser, &mut IoStreams, &mut [&wstr]) -> BuiltinResult;
 
 /// The default prompt for the read command.
 pub const DEFAULT_READ_PROMPT: &wstr =
     L!("set_color green; echo -n read; set_color normal; echo -n \"> \"");
 
-/// Error message on missing argument.
-pub const BUILTIN_ERR_MISSING: &str = "%ls: %ls: option requires an argument\n";
+localizable_consts!(
+    /// Error message on missing argument.
+    pub BUILTIN_ERR_MISSING
+    "%ls: %ls: option requires an argument\n"
 
-/// Error message on missing man page.
-pub const BUILTIN_ERR_MISSING_HELP: &str = concat!(
-    "fish: %ls: missing man page\nDocumentation may not be installed.\n`help %ls` will ",
-    "show an online version\n"
+    /// Error message on unexpected argument.
+    pub BUILTIN_ERR_UNEXP_ARG
+    "%ls: %ls: option does not take an argument\n"
+
+    /// Error message on missing man page.
+    pub BUILTIN_ERR_MISSING_HELP
+    "fish: %ls: missing man page\nDocumentation may not be installed.\n`help %ls` will show an online version\n"
+
+    /// Error message on multiple scope levels for variables.
+    pub BUILTIN_ERR_GLOCAL
+    "%ls: scope can be only one of: universal function global local\n"
+
+    /// Error message for specifying both export and unexport to set/read.
+    pub BUILTIN_ERR_EXPUNEXP
+    "%ls: cannot both export and unexport\n"
+
+    /// Error message for specifying both path and unpath to set/read.
+    pub BUILTIN_ERR_PATHUNPATH
+    "%ls: cannot both path and unpath\n"
+
+    /// Error message for unknown switch.
+    pub BUILTIN_ERR_UNKNOWN
+    "%ls: %ls: unknown option\n"
+
+    /// Error message for invalid bind mode name.
+    pub BUILTIN_ERR_BIND_MODE
+    "%ls: %ls: invalid mode name. See `help identifiers`\n"
+
+    /// Error message when too many arguments are supplied to a builtin.
+    pub BUILTIN_ERR_TOO_MANY_ARGUMENTS
+    "%ls: too many arguments\n"
+
+    /// Error message when integer expected
+    pub BUILTIN_ERR_NOT_NUMBER
+    "%ls: %ls: invalid integer\n"
+
+    /// Command that requires a subcommand was invoked without a recognized subcommand.
+    pub BUILTIN_ERR_MISSING_SUBCMD
+    "%ls: missing subcommand\n"
+
+    pub BUILTIN_ERR_INVALID_SUBCMD
+    "%ls: %ls: invalid subcommand\n"
+
+    /// Error messages for unexpected args.
+    pub BUILTIN_ERR_ARG_COUNT0
+    "%ls: missing argument\n"
+
+    pub BUILTIN_ERR_ARG_COUNT1
+    "%ls: expected %d arguments; got %d\n"
+
+    pub BUILTIN_ERR_ARG_COUNT2
+    "%ls: %ls: expected %d arguments; got %d\n"
+
+    pub BUILTIN_ERR_MIN_ARG_COUNT1
+    "%ls: expected >= %d arguments; got %d\n"
+
+    pub BUILTIN_ERR_MAX_ARG_COUNT1
+    "%ls: expected <= %d arguments; got %d\n"
+
+    /// Error message for invalid variable name.
+    pub BUILTIN_ERR_VARNAME
+    "%ls: %ls: invalid variable name. See `help identifiers`\n"
+
+    /// Error message on invalid combination of options.
+    pub BUILTIN_ERR_COMBO
+    "%ls: invalid option combination\n"
+
+    pub BUILTIN_ERR_COMBO2
+    "%ls: invalid option combination, %ls\n"
+
+    pub BUILTIN_ERR_COMBO2_EXCLUSIVE
+    "%ls: %ls %ls: options cannot be used together\n"
+
+    /// The send stuff to foreground message.
+    pub FG_MSG
+    "Send job %d (%ls) to foreground\n"
 );
 
-/// Error message on multiple scope levels for variables.
-pub const BUILTIN_ERR_GLOCAL: &str =
-    "%ls: scope can be only one of: universal function global local\n";
-
-/// Error message for specifying both export and unexport to set/read.
-pub const BUILTIN_ERR_EXPUNEXP: &str = "%ls: cannot both export and unexport\n";
-
-/// Error message for specifying both path and unpath to set/read.
-pub const BUILTIN_ERR_PATHUNPATH: &str = "%ls: cannot both path and unpath\n";
-
-/// Error message for unknown switch.
-pub const BUILTIN_ERR_UNKNOWN: &str = "%ls: %ls: unknown option\n";
-
-/// Error message for invalid bind mode name.
-pub const BUILTIN_ERR_BIND_MODE: &str = "%ls: %ls: invalid mode name. See `help identifiers`\n";
-
-/// Error message when too many arguments are supplied to a builtin.
-pub const BUILTIN_ERR_TOO_MANY_ARGUMENTS: &str = "%ls: too many arguments\n";
-
-/// Error message when integer expected
-pub const BUILTIN_ERR_NOT_NUMBER: &str = "%ls: %ls: invalid integer\n";
-
-/// Command that requires a subcommand was invoked without a recognized subcommand.
-pub const BUILTIN_ERR_MISSING_SUBCMD: &str = "%ls: missing subcommand\n";
-pub const BUILTIN_ERR_INVALID_SUBCMD: &str = "%ls: %ls: invalid subcommand\n";
-
-/// Error messages for unexpected args.
-pub const BUILTIN_ERR_ARG_COUNT0: &str = "%ls: missing argument\n";
-pub const BUILTIN_ERR_ARG_COUNT1: &str = "%ls: expected %d arguments; got %d\n";
-pub const BUILTIN_ERR_ARG_COUNT2: &str = "%ls: %ls: expected %d arguments; got %d\n";
-pub const BUILTIN_ERR_MIN_ARG_COUNT1: &str = "%ls: expected >= %d arguments; got %d\n";
-pub const BUILTIN_ERR_MAX_ARG_COUNT1: &str = "%ls: expected <= %d arguments; got %d\n";
-
-/// Error message for invalid variable name.
-pub const BUILTIN_ERR_VARNAME: &str = "%ls: %ls: invalid variable name. See `help identifiers`\n";
-
-/// Error message on invalid combination of options.
-pub const BUILTIN_ERR_COMBO: &str = "%ls: invalid option combination\n";
-pub const BUILTIN_ERR_COMBO2: &str = "%ls: invalid option combination, %ls\n";
-pub const BUILTIN_ERR_COMBO2_EXCLUSIVE: &str = "%ls: %ls %ls: options cannot be used together\n";
-
-/// The send stuff to foreground message.
-pub const FG_MSG: &str = "Send job %d (%ls) to foreground\n";
-
 // Return values (`$status` values for fish scripts) for various situations.
 
+#[derive(Clone, Copy, Debug, Eq, PartialEq)]
+pub struct Success {
+    pub preserve_failure_exit_status: bool,
+}
+
+pub const SUCCESS: Success = Success {
+    preserve_failure_exit_status: false,
+};
+
+// To-do: this should be a nonzero type.
+pub type ErrorCode = c_int;
+
+pub type BuiltinResult = Result<Success, ErrorCode>;
+
+pub trait BuiltinResultExt {
+    fn from_dynamic(code: c_int) -> Self;
+    fn builtin_status_code(&self) -> c_int;
+}
+
+impl BuiltinResultExt for BuiltinResult {
+    fn from_dynamic(code: c_int) -> Self {
+        if code == 0 {
+            Ok(SUCCESS)
+        } else {
+            Err(code)
+        }
+    }
+    fn builtin_status_code(&self) -> c_int {
+        match self {
+            Ok(success) => {
+                assert!(!success.preserve_failure_exit_status);
+                0
+            }
+            Err(err) => *err,
+        }
+    }
+}
+
 /// The status code used for normal exit in a command.
-pub const STATUS_CMD_OK: Option<c_int> = Some(0);
+pub const STATUS_CMD_OK: c_int = 0;
+
 /// The status code used for failure exit in a command (but not if the args were invalid).
-pub const STATUS_CMD_ERROR: Option<c_int> = Some(1);
+pub const STATUS_CMD_ERROR: c_int = 1;
 /// The status code used for invalid arguments given to a command. This is distinct from valid
 /// arguments that might result in a command failure. An invalid args condition is something
 /// like an unrecognized flag, missing or too many arguments, an invalid integer, etc.
-pub const STATUS_INVALID_ARGS: Option<c_int> = Some(2);
+pub const STATUS_INVALID_ARGS: c_int = 2;
 
 /// The status code used when a command was not found.
-pub const STATUS_CMD_UNKNOWN: Option<c_int> = Some(127);
+pub const STATUS_CMD_UNKNOWN: c_int = 127;
 
 /// The status code used when an external command can not be run.
-pub const STATUS_NOT_EXECUTABLE: Option<c_int> = Some(126);
+pub const STATUS_NOT_EXECUTABLE: c_int = 126;
 
 /// The status code used when a wildcard had no matches.
-pub const STATUS_UNMATCHED_WILDCARD: Option<c_int> = Some(124);
+pub const STATUS_UNMATCHED_WILDCARD: c_int = 124;
 /// The status code used when illegal command name is encountered.
-pub const STATUS_ILLEGAL_CMD: Option<c_int> = Some(123);
+pub const STATUS_ILLEGAL_CMD: c_int = 123;
 /// The status code used when `read` is asked to consume too much data.
-pub const STATUS_READ_TOO_MUCH: Option<c_int> = Some(122);
+pub const STATUS_READ_TOO_MUCH: c_int = 122;
 /// The status code when an expansion fails, for example, "$foo["
-pub const STATUS_EXPAND_ERROR: Option<c_int> = Some(121);
+pub const STATUS_EXPAND_ERROR: c_int = 121;
+
+pub const STATUS_NO_VARIABLES_GIVEN: c_int = 255;
 
 /// Data structure to describe a builtin.
 struct BuiltinData {
@@ -113,6 +183,10 @@ struct BuiltinData {
 // Functions that are bound to builtin_generic are handled directly by the parser.
 // NOTE: These must be kept in sorted order!
 const BUILTIN_DATAS: &[BuiltinData] = &[
+    BuiltinData {
+        name: L!("!"),
+        func: builtin_generic,
+    },
     BuiltinData {
         name: L!("."),
         func: source::source,
@@ -241,6 +315,14 @@ const BUILTIN_DATAS: &[BuiltinData] = &[
         name: L!("fg"),
         func: fg::fg,
     },
+    BuiltinData {
+        name: L!("fish_indent"),
+        func: fish_indent::fish_indent,
+    },
+    BuiltinData {
+        name: L!("fish_key_reader"),
+        func: fish_key_reader::fish_key_reader,
+    },
     BuiltinData {
         name: L!("for"),
         func: builtin_generic,
@@ -392,7 +474,7 @@ fn cmd_needs_help(cmd: &wstr) -> bool {
 /// Execute a builtin command
 pub fn builtin_run(parser: &Parser, argv: &mut [&wstr], streams: &mut IoStreams) -> ProcStatus {
     if argv.is_empty() {
-        return ProcStatus::from_exit_code(STATUS_INVALID_ARGS.unwrap());
+        return ProcStatus::from_exit_code(STATUS_INVALID_ARGS);
     }
 
     // We can be handed a keyword by the parser as if it was a command. This happens when the user
@@ -400,12 +482,12 @@ pub fn builtin_run(parser: &Parser, argv: &mut [&wstr], streams: &mut IoStreams)
     // handle displaying help for it here.
     if argv.len() == 2 && parse_util_argument_is_help(argv[1]) && cmd_needs_help(argv[0]) {
         builtin_print_help(parser, streams, argv[0]);
-        return ProcStatus::from_exit_code(STATUS_CMD_OK.unwrap());
+        return ProcStatus::from_exit_code(STATUS_CMD_OK);
     }
 
     let Some(builtin) = builtin_lookup(argv[0]) else {
         FLOGF!(error, "%s", wgettext_fmt!(UNKNOWN_BUILTIN_ERR_MSG, argv[0]));
-        return ProcStatus::from_exit_code(STATUS_CMD_ERROR.unwrap());
+        return ProcStatus::from_exit_code(STATUS_CMD_ERROR);
     };
 
     let builtin_ret = (builtin.func)(parser, streams, argv);
@@ -417,7 +499,17 @@ pub fn builtin_run(parser: &Parser, argv: &mut [&wstr], streams: &mut IoStreams)
     // Resolve our status code.
     // If the builtin itself produced an error, use that error.
     // Otherwise use any errors from writing to out and writing to err, in that order.
-    let mut code = builtin_ret.unwrap_or(0);
+    let mut code = match builtin_ret {
+        Ok(success) => {
+            if success.preserve_failure_exit_status {
+                return ProcStatus::empty();
+            } else {
+                0
+            }
+        }
+        Err(err) => err,
+    };
+
     if code == 0 {
         code = out_ret;
     }
@@ -431,10 +523,6 @@ pub fn builtin_run(parser: &Parser, argv: &mut [&wstr], streams: &mut IoStreams)
         code = 255;
     }
 
-    // Handle the case of an empty status.
-    if code == 0 && builtin_ret.is_none() {
-        return ProcStatus::empty();
-    }
     if code < 0 {
         // If the code is below 0, constructing a proc_status_t
         // would assert() out, which is a terrible failure mode
@@ -463,6 +551,7 @@ pub fn builtin_get_names() -> impl Iterator<Item = &'static wstr> {
 /// Return a one-line description of the specified builtin.
 pub fn builtin_get_desc(name: &wstr) -> Option<&'static wstr> {
     let desc = match name {
+        _ if name == "!" => wgettext!("Negate exit status of job"),
         _ if name == "." => wgettext!("Evaluate contents of file"),
         _ if name == ":" => wgettext!("Return a successful result"),
         _ if name == "[" => wgettext!("Test a condition"), // ]
@@ -495,6 +584,7 @@ pub fn builtin_get_desc(name: &wstr) -> Option<&'static wstr> {
         _ if name == "exit" => wgettext!("Exit the shell"),
         _ if name == "false" => wgettext!("Return an unsuccessful result"),
         _ if name == "fg" => wgettext!("Send job to foreground"),
+        _ if name == "fish_key_reader" => wgettext!("explore what characters keyboard keys send"),
         _ if name == "for" => wgettext!("Perform a set of commands multiple times"),
         _ if name == "function" => wgettext!("Define a new function"),
         _ if name == "functions" => wgettext!("List or remove functions"),
@@ -536,28 +626,13 @@ pub fn builtin_get_desc(name: &wstr) -> Option<&'static wstr> {
 ///
 /// Process and print help for the specified builtin or function.
 pub fn builtin_print_help(parser: &Parser, streams: &mut IoStreams, cmd: &wstr) {
-    builtin_print_help_error(parser, streams, cmd, L!(""))
-}
-
-pub fn builtin_print_help_error(
-    parser: &Parser,
-    streams: &mut IoStreams,
-    cmd: &wstr,
-    error_message: &wstr,
-) {
     // This won't ever work if no_exec is set.
     if no_exec() {
         return;
     }
     let name_esc = escape(cmd);
-    let mut cmd = sprintf!("__fish_print_help %ls ", &name_esc);
-    let mut ios = streams.io_chain.clone();
-    if !error_message.is_empty() {
-        cmd.push_utfstr(&escape(error_message));
-        // If it's an error, redirect the output of __fish_print_help to stderr
-        ios.push(Arc::new(IoFd::new(STDOUT_FILENO, STDERR_FILENO)));
-    }
-    let res = parser.eval(&cmd, &ios);
+    let cmd = sprintf!("__fish_print_help %ls ", &name_esc);
+    let res = parser.eval(&cmd, streams.io_chain);
     if res.status.normal_exited() && res.status.exit_code() == 2 {
         streams
             .err
@@ -607,6 +682,22 @@ pub fn builtin_missing_argument(
     }
 }
 
+/// Perform error reporting for encounter with an extra argument.
+pub fn builtin_unexpected_argument(
+    parser: &Parser,
+    streams: &mut IoStreams,
+    cmd: &wstr,
+    opt: &wstr,
+    print_hints: bool, /*=true*/
+) {
+    streams
+        .err
+        .append(wgettext_fmt!(BUILTIN_ERR_UNEXP_ARG, cmd, opt));
+    if print_hints {
+        builtin_print_error_trailer(parser, streams.err, cmd);
+    }
+}
+
 /// Print the backtrace and call for help that we use at the end of error messages.
 pub fn builtin_print_error_trailer(parser: &Parser, b: &mut OutputStream, cmd: &wstr) {
     b.push('\n');
@@ -645,11 +736,11 @@ impl HelpOnlyCmdOpts {
         args: &mut [&wstr],
         parser: &Parser,
         streams: &mut IoStreams,
-    ) -> Result<Self, Option<c_int>> {
+    ) -> Result<Self, ErrorCode> {
         let cmd = args[0];
         let print_hints = true;
 
-        const shortopts: &wstr = L!("+:h");
+        const shortopts: &wstr = L!("+h");
         const longopts: &[WOption] = &[wopt(L!("help"), ArgType::NoArgument, 'h')];
 
         let mut print_help = false;
@@ -669,6 +760,16 @@ impl HelpOnlyCmdOpts {
                     );
                     return Err(STATUS_INVALID_ARGS);
                 }
+                ';' => {
+                    builtin_unexpected_argument(
+                        parser,
+                        streams,
+                        cmd,
+                        args[w.wopt_index - 1],
+                        print_hints,
+                    );
+                    return Err(STATUS_INVALID_ARGS);
+                }
                 '?' => {
                     builtin_unknown_option(
                         parser,
@@ -832,26 +933,23 @@ impl<'args> Iterator for Arguments<'args, '_> {
 
 /// A generic builtin that only supports showing a help message. This is only a placeholder that
 /// prints the help message. Useful for commands that live in the parser.
-fn builtin_generic(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Option<c_int> {
+fn builtin_generic(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> BuiltinResult {
     let argc = argv.len();
-    let opts = match HelpOnlyCmdOpts::parse(argv, parser, streams) {
-        Ok(opts) => opts,
-        Err(err) => return err,
-    };
+    let opts = HelpOnlyCmdOpts::parse(argv, parser, streams)?;
 
     if opts.print_help {
         builtin_print_help(parser, streams, argv[0]);
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     }
 
     // Hackish - if we have no arguments other than the command, we are a "naked invocation" and we
     // just print help.
     if argc == 1 || argv[0] == "time" {
         builtin_print_help(parser, streams, argv[0]);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
-    STATUS_CMD_ERROR
+    Err(STATUS_CMD_ERROR)
 }
 
 /// This function handles both the 'continue' and the 'break' builtins that are used for loop
@@ -860,20 +958,28 @@ fn builtin_break_continue(
     parser: &Parser,
     streams: &mut IoStreams,
     argv: &mut [&wstr],
-) -> Option<c_int> {
+) -> BuiltinResult {
     let is_break = argv[0] == "break";
     let argc = argv.len();
 
+    let opts = HelpOnlyCmdOpts::parse(argv, parser, streams)?;
+
+    if opts.print_help {
+        builtin_print_help(parser, streams, argv[0]);
+        return Ok(SUCCESS);
+    }
+
     if argc != 1 {
-        let error_message = wgettext_fmt!(BUILTIN_ERR_UNKNOWN, argv[0], argv[1]);
-        builtin_print_help_error(parser, streams, argv[0], &error_message);
-        return STATUS_INVALID_ARGS;
+        streams
+            .err
+            .append(wgettext_fmt!(BUILTIN_ERR_UNKNOWN, argv[0], argv[1]));
+        return Err(STATUS_INVALID_ARGS);
     }
 
     // Paranoia: ensure we have a real loop.
     // This is checked in the AST but we may be invoked dynamically, e.g. just via "eval break".
     let mut has_loop = false;
-    for b in parser.blocks().iter().rev() {
+    for b in parser.blocks_iter_rev() {
         if [BlockType::while_block, BlockType::for_block].contains(&b.typ()) {
             has_loop = true;
             break;
@@ -883,9 +989,10 @@ fn builtin_break_continue(
         }
     }
     if !has_loop {
-        let error_message = wgettext_fmt!("%ls: Not inside of loop\n", argv[0]);
-        builtin_print_help_error(parser, streams, argv[0], &error_message);
-        return STATUS_CMD_ERROR;
+        streams
+            .err
+            .append(wgettext_fmt!("%ls: Not inside of loop\n", argv[0]));
+        return Err(STATUS_CMD_ERROR);
     }
 
     // Mark the status in the libdata.
@@ -894,7 +1001,7 @@ fn builtin_break_continue(
     } else {
         LoopStatus::continues
     };
-    STATUS_CMD_OK
+    Ok(SUCCESS)
 }
 
 /// Implementation of the builtin breakpoint command, used to launch the interactive debugger.
@@ -902,7 +1009,7 @@ fn builtin_breakpoint(
     parser: &Parser,
     streams: &mut IoStreams,
     argv: &mut [&wstr],
-) -> Option<c_int> {
+) -> BuiltinResult {
     let cmd = argv[0];
     if argv.len() != 1 {
         streams.err.append(wgettext_fmt!(
@@ -911,12 +1018,12 @@ fn builtin_breakpoint(
             0,
             argv.len() - 1
         ));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     // If we're not interactive then we can't enter the debugger. So treat this command as a no-op.
     if !parser.is_interactive() {
-        return STATUS_CMD_ERROR;
+        return Err(STATUS_CMD_ERROR);
     }
 
     // Ensure we don't allow creating a breakpoint at an interactive prompt. There may be a simpler
@@ -930,28 +1037,34 @@ fn builtin_breakpoint(
                 "%ls: Command not valid at an interactive prompt\n",
                 cmd,
             ));
-            return STATUS_ILLEGAL_CMD;
+            return Err(STATUS_ILLEGAL_CMD);
         }
     }
 
     let bpb = parser.push_block(Block::breakpoint_block());
     let io_chain = &streams.io_chain;
-    reader_read(parser, STDIN_FILENO, io_chain);
+    reader_read(parser, STDIN_FILENO, io_chain)?;
     parser.pop_block(bpb);
-    Some(parser.get_last_status())
+    BuiltinResult::from_dynamic(parser.get_last_status())
 }
 
-fn builtin_true(_parser: &Parser, _streams: &mut IoStreams, _argv: &mut [&wstr]) -> Option<c_int> {
-    STATUS_CMD_OK
+fn builtin_true(_parser: &Parser, _streams: &mut IoStreams, _argv: &mut [&wstr]) -> BuiltinResult {
+    Ok(SUCCESS)
 }
 
-fn builtin_false(_parser: &Parser, _streams: &mut IoStreams, _argv: &mut [&wstr]) -> Option<c_int> {
-    STATUS_CMD_ERROR
+fn builtin_false(_parser: &Parser, _streams: &mut IoStreams, _argv: &mut [&wstr]) -> BuiltinResult {
+    Err(STATUS_CMD_ERROR)
 }
 
-fn builtin_gettext(_parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Option<c_int> {
+/// Used for the fish `_` builtin for requesting translations.
+/// For scripts in `share/`, the corresponding strings are extracted from the scripts using
+/// `build_tools/fish_xgettext.fish`.
+/// Strings not present in our repo would require a custom MO file for translation to be possible.
+fn builtin_gettext(_parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> BuiltinResult {
     for arg in &argv[1..] {
-        streams.out.append(wgettext_str(arg));
+        streams.out.append(
+            crate::wutil::LocalizableString::from_external_source((*arg).to_owned()).localize(),
+        );
     }
-    STATUS_CMD_OK
+    Ok(SUCCESS)
 }
