@@ -7,6 +7,7 @@ use crate::proc::{
     get_job_control_mode, get_login, is_interactive_session, set_job_control_mode, JobControl,
 };
 use crate::reader::reader_in_interactive_read;
+use crate::tty_handoff::{get_scroll_content_up_capability, xtversion};
 use crate::wutil::{waccess, wbasename, wdirname, wrealpath, Error};
 use libc::F_OK;
 use nix::errno::Errno;
@@ -59,13 +60,19 @@ enum StatusCmd {
     STATUS_STACK_TRACE,
     STATUS_TEST_FEATURE,
     STATUS_CURRENT_COMMANDLINE,
+    STATUS_BUILD_INFO,
     STATUS_BUILDINFO,
+    STATUS_GET_FILE,
+    STATUS_LIST_FILES,
+    STATUS_TERMINAL,
+    STATUS_TEST_TERMINAL_FEATURE,
 }
 
 str_enum!(
     StatusCmd,
     (STATUS_BASENAME, "basename"),
     (STATUS_BASENAME, "current-basename"),
+    (STATUS_BUILD_INFO, "build-info"),
     (STATUS_BUILDINFO, "buildinfo"),
     (STATUS_CURRENT_CMD, "current-command"),
     (STATUS_CURRENT_COMMANDLINE, "current-commandline"),
@@ -78,6 +85,8 @@ str_enum!(
     (STATUS_FILENAME, "filename"),
     (STATUS_FISH_PATH, "fish-path"),
     (STATUS_FUNCTION, "function"),
+    (STATUS_GET_FILE, "get-file"),
+    (STATUS_LIST_FILES, "list-files"),
     (STATUS_IS_BLOCK, "is-block"),
     (STATUS_IS_BREAKPOINT, "is-breakpoint"),
     (STATUS_IS_COMMAND_SUB, "is-command-substitution"),
@@ -91,7 +100,9 @@ str_enum!(
     (STATUS_LINE_NUMBER, "line-number"),
     (STATUS_STACK_TRACE, "print-stack-trace"),
     (STATUS_STACK_TRACE, "stack-trace"),
+    (STATUS_TERMINAL, "terminal"),
     (STATUS_TEST_FEATURE, "test-feature"),
+    (STATUS_TEST_TERMINAL_FEATURE, "test-terminal-feature"),
 );
 
 /// Values that may be returned from the test-feature option to status.
@@ -146,7 +157,7 @@ const IS_INTERACTIVE_JOB_CTRL_SHORT: char = '\x03';
 const IS_NO_JOB_CTRL_SHORT: char = '\x04';
 const IS_INTERACTIVE_READ_SHORT: char = '\x05';
 
-const SHORT_OPTIONS: &wstr = L!(":L:cbilfnhj:t");
+const SHORT_OPTIONS: &wstr = L!("L:cbilfnhj:t");
 const LONG_OPTIONS: &[WOption] = &[
     wopt(L!("help"), NoArgument, 'h'),
     wopt(L!("current-filename"), NoArgument, 'f'),
@@ -210,7 +221,7 @@ fn parse_cmd_opts(
     args: &mut [&wstr],
     parser: &Parser,
     streams: &mut IoStreams,
-) -> Option<c_int> {
+) -> BuiltinResult {
     let cmd = args[0];
 
     let mut args_read = Vec::with_capacity(args.len());
@@ -230,13 +241,13 @@ fn parse_cmd_opts(
                                 cmd,
                                 arg
                             ));
-                            return STATUS_INVALID_ARGS;
+                            return Err(STATUS_INVALID_ARGS);
                         }
                         _ => {
                             streams
                                 .err
                                 .append(wgettext_fmt!(BUILTIN_ERR_NOT_NUMBER, cmd, arg));
-                            return STATUS_INVALID_ARGS;
+                            return Err(STATUS_INVALID_ARGS);
                         }
                     }
                 };
@@ -253,12 +264,12 @@ fn parse_cmd_opts(
                     _ => unreachable!(),
                 };
                 if !opts.try_set_status_cmd(subcmd, streams) {
-                    return STATUS_CMD_ERROR;
+                    return Err(STATUS_CMD_ERROR);
                 }
             }
             'j' => {
                 if !opts.try_set_status_cmd(STATUS_SET_JOB_CONTROL, streams) {
-                    return STATUS_CMD_ERROR;
+                    return Err(STATUS_CMD_ERROR);
                 }
                 let Ok(job_mode) = w.woptarg.unwrap().try_into() else {
                     streams.err.append(wgettext_fmt!(
@@ -266,43 +277,47 @@ fn parse_cmd_opts(
                         cmd,
                         w.woptarg.unwrap()
                     ));
-                    return STATUS_CMD_ERROR;
+                    return Err(STATUS_CMD_ERROR);
                 };
                 opts.new_job_control_mode = Some(job_mode);
             }
             IS_FULL_JOB_CTRL_SHORT => {
                 if !opts.try_set_status_cmd(STATUS_IS_FULL_JOB_CTRL, streams) {
-                    return STATUS_CMD_ERROR;
+                    return Err(STATUS_CMD_ERROR);
                 }
             }
             IS_INTERACTIVE_JOB_CTRL_SHORT => {
                 if !opts.try_set_status_cmd(STATUS_IS_INTERACTIVE_JOB_CTRL, streams) {
-                    return STATUS_CMD_ERROR;
+                    return Err(STATUS_CMD_ERROR);
                 }
             }
             IS_INTERACTIVE_READ_SHORT => {
                 if !opts.try_set_status_cmd(STATUS_IS_INTERACTIVE_READ, streams) {
-                    return STATUS_CMD_ERROR;
+                    return Err(STATUS_CMD_ERROR);
                 }
             }
             IS_NO_JOB_CTRL_SHORT => {
                 if !opts.try_set_status_cmd(STATUS_IS_NO_JOB_CTRL, streams) {
-                    return STATUS_CMD_ERROR;
+                    return Err(STATUS_CMD_ERROR);
                 }
             }
             FISH_PATH_SHORT => {
                 if !opts.try_set_status_cmd(STATUS_FISH_PATH, streams) {
-                    return STATUS_CMD_ERROR;
+                    return Err(STATUS_CMD_ERROR);
                 }
             }
             'h' => opts.print_help = true,
             ':' => {
                 builtin_missing_argument(parser, streams, cmd, args[w.wopt_index - 1], false);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
+            }
+            ';' => {
+                builtin_unexpected_argument(parser, streams, cmd, args[w.wopt_index - 1], false);
+                return Err(STATUS_INVALID_ARGS);
             }
             '?' => {
                 builtin_unknown_option(parser, streams, cmd, args[w.wopt_index - 1], false);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
             _ => panic!("unexpected retval from WGetopter"),
         }
@@ -310,23 +325,34 @@ fn parse_cmd_opts(
 
     *optind = w.wopt_index;
 
-    return STATUS_CMD_OK;
+    return Ok(SUCCESS);
 }
 
-pub fn status(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
+#[cfg(feature = "embed-data")]
+use rust_embed::RustEmbed;
+
+#[cfg(feature = "embed-data")]
+#[derive(RustEmbed)]
+#[folder = "target/fish-man/man1"]
+#[prefix = "man/man1/"]
+struct Docs;
+
+pub fn status(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
+    localizable_consts!(
+        #[allow(dead_code)]
+        NO_EMBEDDED_FILES_MSG "%ls: fish was not built with embedded files"
+    );
+
     let cmd = args[0];
     let argc = args.len();
 
     let mut opts = StatusCmdOpts::default();
     let mut optind = 0usize;
-    let retval = parse_cmd_opts(&mut opts, &mut optind, args, parser, streams);
-    if retval != STATUS_CMD_OK {
-        return retval;
-    }
+    parse_cmd_opts(&mut opts, &mut optind, args, parser, streams)?;
 
     if opts.print_help {
         builtin_print_help(parser, streams, cmd);
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     }
 
     // If a status command hasn't already been specified via a flag check the first word.
@@ -335,7 +361,7 @@ pub fn status(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> O
         match StatusCmd::from_wstr(args[optind].to_string().as_str()) {
             Some(s) => {
                 if !opts.try_set_status_cmd(s, streams) {
-                    return STATUS_CMD_ERROR;
+                    return Err(STATUS_CMD_ERROR);
                 }
                 optind += 1;
             }
@@ -343,7 +369,7 @@ pub fn status(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> O
                 streams
                     .err
                     .append(wgettext_fmt!(BUILTIN_ERR_INVALID_SUBCMD, cmd, args[1]));
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
         }
     }
@@ -359,16 +385,16 @@ pub fn status(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> O
             streams.out.append(wgettext!("This is not a login shell\n"));
         }
         let job_control_mode = match get_job_control_mode() {
-            JobControl::interactive => wgettext!("Only on interactive jobs"),
-            JobControl::none => wgettext!("Never"),
-            JobControl::all => wgettext!("Always"),
+            JobControl::Interactive => wgettext!("Only on interactive jobs"),
+            JobControl::None => wgettext!("Never"),
+            JobControl::All => wgettext!("Always"),
         };
         streams
             .out
             .append(wgettext_fmt!("Job control: %ls\n", job_control_mode));
         streams.out.append(parser.stack_trace());
 
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     };
 
     match subcmd {
@@ -384,7 +410,7 @@ pub fn status(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> O
                             0,
                             args.len()
                         ));
-                        return STATUS_INVALID_ARGS;
+                        return Err(STATUS_INVALID_ARGS);
                     }
                     j
                 }
@@ -397,7 +423,7 @@ pub fn status(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> O
                             1,
                             args.len()
                         ));
-                        return STATUS_INVALID_ARGS;
+                        return Err(STATUS_INVALID_ARGS);
                     }
                     let Ok(new_mode) = args[0].try_into() else {
                         streams.err.append(wgettext_fmt!(
@@ -405,7 +431,7 @@ pub fn status(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> O
                             cmd,
                             args[0]
                         ));
-                        return STATUS_CMD_ERROR;
+                        return Err(STATUS_CMD_ERROR);
                     };
                     new_mode
                 }
@@ -422,59 +448,126 @@ pub fn status(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> O
                     1,
                     args.len()
                 ));
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
-            use TestFeatureRetVal::*;
-            let mut retval = Some(TEST_FEATURE_NOT_RECOGNIZED as c_int);
+            let mut retval = TestFeatureRetVal::TEST_FEATURE_NOT_RECOGNIZED;
             for md in features::METADATA {
                 if md.name == args[0] {
                     retval = match feature_test(md.flag) {
-                        true => Some(TEST_FEATURE_ON as c_int),
-                        false => Some(TEST_FEATURE_OFF as c_int),
+                        true => TestFeatureRetVal::TEST_FEATURE_ON,
+                        false => TestFeatureRetVal::TEST_FEATURE_OFF,
                     };
                 }
             }
-            return retval;
+            return Err(retval as i32);
         }
-        STATUS_BUILDINFO => {
-            let version = str2wcstring(crate::BUILD_VERSION.as_bytes());
-            let target = str2wcstring(env!("BUILD_TARGET_TRIPLE").as_bytes());
-            let host = str2wcstring(env!("BUILD_HOST_TRIPLE").as_bytes());
-            let profile = str2wcstring(env!("BUILD_PROFILE").as_bytes());
-            streams.out.append(L!("Build system: "));
-            let buildsystem = match option_env!("CMAKE") {
-                Some("1") => "CMake",
-                _ => "Cargo",
-            };
-            streams.out.appendln(str2wcstring(buildsystem.as_bytes()));
-            streams.out.append(L!("Version: "));
-            streams.out.appendln(version);
-            if target == host {
-                streams.out.append(L!("Target (and host): "));
-                streams.out.appendln(target);
-            } else {
-                streams.out.append(L!("Target: "));
-                streams.out.appendln(target);
-                streams.out.append(L!("Host: "));
-                streams.out.appendln(host);
+        c @ STATUS_GET_FILE => {
+            if args.len() != 1 {
+                streams.err.append(wgettext_fmt!(
+                    BUILTIN_ERR_ARG_COUNT2,
+                    cmd,
+                    c.to_wstr(),
+                    1,
+                    args.len()
+                ));
+                return Err(STATUS_INVALID_ARGS);
+            }
+            #[cfg(feature = "embed-data")]
+            {
+                let arg = crate::common::wcs2string(args[0]);
+                let arg = std::str::from_utf8(&arg).unwrap();
+                if let Some(emfile) = crate::autoload::Asset::get(arg) {
+                    let src = str2wcstring(&emfile.data);
+                    streams.out.append(src);
+                    return Ok(SUCCESS);
+                } else if let Some(emfile) = Docs::get(arg) {
+                    let src = str2wcstring(&emfile.data);
+                    streams.out.append(src);
+                    return Ok(SUCCESS);
+                } else {
+                    return Err(STATUS_CMD_ERROR);
+                }
+            }
+            #[cfg(not(feature = "embed-data"))]
+            {
+                streams
+                    .err
+                    .appendln(sprintf!(NO_EMBEDDED_FILES_MSG.localize(), cmd));
+                return Err(STATUS_CMD_ERROR);
             }
-            streams.out.append(L!("Profile: "));
-            streams.out.appendln(profile);
-            streams.out.append(L!("Features: "));
-            let features: &[&str] = &[
-                #[cfg(gettext)]
-                "gettext",
-                #[cfg(feature = "installable")]
-                "installable",
-                #[cfg(target_feature = "crt-static")]
-                "crt-static",
-            ];
-            streams
-                .out
-                .appendln(str2wcstring(features.join(" ").as_bytes()));
-            streams.out.appendln("");
-            return STATUS_CMD_OK;
         }
+        c @ STATUS_LIST_FILES => {
+            if args.len() > 1 {
+                streams.err.append(wgettext_fmt!(
+                    BUILTIN_ERR_ARG_COUNT2,
+                    cmd,
+                    c.to_wstr(),
+                    1,
+                    args.len()
+                ));
+                return Err(STATUS_INVALID_ARGS);
+            }
+            #[cfg(feature = "embed-data")]
+            {
+                let mut have_file = false;
+                let arg = crate::common::wcs2string(args.get(0).unwrap_or(&L!("")));
+                let arg = std::str::from_utf8(&arg).unwrap();
+                for file in crate::autoload::Asset::iter() {
+                    if arg.is_empty() || file.starts_with(arg) {
+                        have_file = true;
+                        let src = str2wcstring(file.as_bytes());
+                        streams.out.appendln(src);
+                    }
+                }
+                for file in Docs::iter() {
+                    if arg.is_empty() || file.starts_with(arg) {
+                        have_file = true;
+                        let src = str2wcstring(file.as_bytes());
+                        streams.out.appendln(src);
+                    }
+                }
+
+                if have_file {
+                    return Ok(SUCCESS);
+                } else {
+                    return Err(STATUS_CMD_ERROR);
+                }
+            }
+            #[cfg(not(feature = "embed-data"))]
+            {
+                streams
+                    .err
+                    .appendln(sprintf!(NO_EMBEDDED_FILES_MSG.localize(), cmd));
+                return Err(STATUS_CMD_ERROR);
+            }
+        }
+        c @ STATUS_TEST_TERMINAL_FEATURE => {
+            if args.len() != 1 {
+                streams.err.append(wgettext_fmt!(
+                    BUILTIN_ERR_ARG_COUNT2,
+                    cmd,
+                    c.to_wstr(),
+                    1,
+                    args.len()
+                ));
+                return Err(STATUS_INVALID_ARGS);
+            }
+            if args[0] != "scroll-content-up" {
+                streams.err.appendln(wgettext_fmt!(
+                    "%s %s: unrecognized feature '%ls'",
+                    cmd,
+                    c.to_wstr(),
+                    args[0]
+                ));
+                return Err(STATUS_INVALID_ARGS);
+            };
+            return if get_scroll_content_up_capability() == Some(true) {
+                Ok(SUCCESS)
+            } else {
+                Err(STATUS_CMD_ERROR)
+            };
+        }
+
         ref s => {
             if !args.is_empty() {
                 streams.err.append(wgettext_fmt!(
@@ -484,9 +577,48 @@ pub fn status(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> O
                     0,
                     args.len()
                 ));
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
             match s {
+                STATUS_BUILD_INFO | STATUS_BUILDINFO => {
+                    let version = str2wcstring(crate::BUILD_VERSION.as_bytes());
+                    let target = str2wcstring(env!("BUILD_TARGET_TRIPLE").as_bytes());
+                    let host = str2wcstring(env!("BUILD_HOST_TRIPLE").as_bytes());
+                    let profile = str2wcstring(env!("BUILD_PROFILE").as_bytes());
+                    streams.out.append(L!("Build system: "));
+                    let buildsystem = match option_env!("CMAKE") {
+                        Some("1") => "CMake",
+                        _ => "Cargo",
+                    };
+                    streams.out.appendln(str2wcstring(buildsystem.as_bytes()));
+                    streams.out.append(L!("Version: "));
+                    streams.out.appendln(version);
+                    if target == host {
+                        streams.out.append(L!("Target (and host): "));
+                        streams.out.appendln(target);
+                    } else {
+                        streams.out.append(L!("Target: "));
+                        streams.out.appendln(target);
+                        streams.out.append(L!("Host: "));
+                        streams.out.appendln(host);
+                    }
+                    streams.out.append(L!("Profile: "));
+                    streams.out.appendln(profile);
+                    streams.out.append(L!("Features: "));
+                    let features: &[&str] = &[
+                        #[cfg(feature = "embed-data")]
+                        "embed-data",
+                        #[cfg(feature = "localize-messages")]
+                        "localize-messages",
+                        #[cfg(target_feature = "crt-static")]
+                        "crt-static",
+                    ];
+                    streams
+                        .out
+                        .appendln(str2wcstring(features.join(" ").as_bytes()));
+                    streams.out.appendln("");
+                    return Ok(SUCCESS);
+                }
                 STATUS_BASENAME | STATUS_DIRNAME | STATUS_FILENAME => {
                     let res = parser.current_filename();
                     let function = res.unwrap_or_default();
@@ -515,65 +647,64 @@ pub fn status(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> O
                 }
                 STATUS_IS_INTERACTIVE => {
                     if is_interactive_session() {
-                        return STATUS_CMD_OK;
-                    } else {
-                        return STATUS_CMD_ERROR;
+                        return Ok(SUCCESS);
                     }
+                    return Err(STATUS_CMD_ERROR);
                 }
                 STATUS_IS_COMMAND_SUB => {
-                    if parser.libdata().is_subshell {
-                        return STATUS_CMD_OK;
+                    if parser.scope().is_subshell {
+                        return Ok(SUCCESS);
                     } else {
-                        return STATUS_CMD_ERROR;
+                        return Err(STATUS_CMD_ERROR);
                     }
                 }
                 STATUS_IS_BLOCK => {
                     if parser.is_block() {
-                        return STATUS_CMD_OK;
+                        return Ok(SUCCESS);
                     } else {
-                        return STATUS_CMD_ERROR;
+                        return Err(STATUS_CMD_ERROR);
                     }
                 }
                 STATUS_IS_BREAKPOINT => {
                     if parser.is_breakpoint() {
-                        return STATUS_CMD_OK;
+                        return Ok(SUCCESS);
                     } else {
-                        return STATUS_CMD_ERROR;
+                        return Err(STATUS_CMD_ERROR);
                     }
                 }
                 STATUS_IS_LOGIN => {
                     if get_login() {
-                        return STATUS_CMD_OK;
+                        return Ok(SUCCESS);
                     } else {
-                        return STATUS_CMD_ERROR;
+                        return Err(STATUS_CMD_ERROR);
                     }
                 }
                 STATUS_IS_FULL_JOB_CTRL => {
-                    if get_job_control_mode() == JobControl::all {
-                        return STATUS_CMD_OK;
+                    if get_job_control_mode() == JobControl::All {
+                        return Ok(SUCCESS);
                     } else {
-                        return STATUS_CMD_ERROR;
+                        return Err(STATUS_CMD_ERROR);
                     }
                 }
                 STATUS_IS_INTERACTIVE_JOB_CTRL => {
-                    if get_job_control_mode() == JobControl::interactive {
-                        return STATUS_CMD_OK;
+                    if get_job_control_mode() == JobControl::Interactive {
+                        return Ok(SUCCESS);
                     } else {
-                        return STATUS_CMD_ERROR;
+                        return Err(STATUS_CMD_ERROR);
                     }
                 }
                 STATUS_IS_INTERACTIVE_READ => {
                     if reader_in_interactive_read() {
-                        return STATUS_CMD_OK;
+                        return Ok(SUCCESS);
                     } else {
-                        return STATUS_CMD_ERROR;
+                        return Err(STATUS_CMD_ERROR);
                     }
                 }
                 STATUS_IS_NO_JOB_CTRL => {
-                    if get_job_control_mode() == JobControl::none {
-                        return STATUS_CMD_OK;
+                    if get_job_control_mode() == JobControl::None {
+                        return Ok(SUCCESS);
                     } else {
-                        return STATUS_CMD_ERROR;
+                        return Err(STATUS_CMD_ERROR);
                     }
                 }
                 STATUS_STACK_TRACE => {
@@ -603,7 +734,7 @@ pub fn status(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> O
                     }
                     if path.is_absolute() {
                         let path = str2wcstring(path.as_os_str().as_bytes());
-                        // This is an absoulte path, we can canonicalize it
+                        // This is an absolute path, we can canonicalize it
                         let real = match wrealpath(&path) {
                             Some(p) if waccess(&p, F_OK) == 0 => p,
                             // realpath did not work, just append the path
@@ -619,15 +750,25 @@ pub fn status(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> O
                         streams.out.appendln(path);
                     }
                 }
-                STATUS_BUILDINFO
-                | STATUS_SET_JOB_CONTROL
+                STATUS_TERMINAL => {
+                    let xtversion = xtversion().unwrap_or_default();
+                    let first_line = &xtversion[..xtversion
+                        .chars()
+                        .position(|c| c == '\n')
+                        .unwrap_or(xtversion.len())];
+                    streams.out.appendln(first_line);
+                }
+                STATUS_SET_JOB_CONTROL
                 | STATUS_FEATURES
-                | STATUS_TEST_FEATURE => {
+                | STATUS_TEST_FEATURE
+                | STATUS_GET_FILE
+                | STATUS_LIST_FILES
+                | STATUS_TEST_TERMINAL_FEATURE => {
                     unreachable!("")
                 }
             }
         }
     };
 
-    return retval;
+    Ok(SUCCESS)
 }
