@@ -1,21 +1,21 @@
-use crate::common::ToCString;
 use crate::complete::complete_invalidate_path;
-use crate::curses::{self, Term};
 use crate::env::{setenv_lock, unsetenv_lock, EnvMode, EnvStack, Environment};
-use crate::env::{CURSES_INITIALIZED, DEFAULT_READ_BYTE_LIMIT, READ_BYTE_LIMIT, TERM_HAS_XN};
+use crate::env::{DEFAULT_READ_BYTE_LIMIT, READ_BYTE_LIMIT};
 use crate::flog::FLOG;
-use crate::function;
 use crate::input_common::{update_wait_on_escape_ms, update_wait_on_sequence_key_ms};
-use crate::output::ColorSupport;
-use crate::proc::is_interactive_session;
 use crate::reader::{
     reader_change_cursor_end_mode, reader_change_cursor_selection_mode, reader_change_history,
-    reader_schedule_prompt_repaint, reader_set_autosuggestion_enabled,
+    reader_schedule_prompt_repaint, reader_set_autosuggestion_enabled, reader_set_transient_prompt,
 };
-use crate::screen::screen_set_midnight_commander_hack;
-use crate::screen::LAYOUT_CACHE_SHARED;
+use crate::screen::{
+    screen_set_midnight_commander_hack, IS_DUMB, LAYOUT_CACHE_SHARED, ONLY_GRAYSCALE,
+};
+use crate::terminal::use_terminfo;
+use crate::terminal::ColorSupport;
+use crate::tty_handoff::xtversion;
 use crate::wchar::prelude::*;
 use crate::wutil::fish_wcstoi;
+use crate::{function, terminal};
 use std::borrow::Cow;
 use std::collections::HashMap;
 use std::ffi::{CStr, CString};
@@ -50,10 +50,11 @@ static VAR_DISPATCH_TABLE: once_cell::sync::Lazy<VarDispatchTable> =
         }
 
         for name in CURSES_VARIABLES {
-            table.add_anon(name, handle_curses_change);
+            table.add_anon(name, handle_term_change);
         }
 
         table.add(L!("TZ"), handle_tz_change);
+        table.add_anon(L!("COLORTERM"), handle_fish_term_change);
         table.add_anon(L!("fish_term256"), handle_fish_term_change);
         table.add_anon(L!("fish_term24bit"), handle_fish_term_change);
         table.add_anon(L!("fish_escape_delay_ms"), update_wait_on_escape_ms);
@@ -73,6 +74,7 @@ static VAR_DISPATCH_TABLE: once_cell::sync::Lazy<VarDispatchTable> =
             L!("fish_autosuggestion_enabled"),
             handle_autosuggestion_change,
         );
+        table.add_anon(L!("fish_transient_prompt"), handle_transient_prompt_change);
         table.add_anon(
             L!("fish_use_posix_spawn"),
             handle_fish_use_posix_spawn_change,
@@ -151,14 +153,14 @@ fn handle_timezone(var_name: &wstr, vars: &EnvStack) {
     }
 }
 
-/// Update the value of [`FISH_EMOJI_WIDTH`].
-fn guess_emoji_width(vars: &EnvStack) {
+/// Update the value of [`FISH_EMOJI_WIDTH`](crate::fallback::FISH_EMOJI_WIDTH).
+pub fn guess_emoji_width(vars: &EnvStack) {
     use crate::fallback::FISH_EMOJI_WIDTH;
 
     if let Some(width_str) = vars.get(L!("fish_emoji_width")) {
         // The only valid values are 1 or 2; we default to 1 if it was an invalid int.
-        let new_width = fish_wcstoi(&width_str.as_string()).unwrap_or(1).clamp(1, 2);
-        FISH_EMOJI_WIDTH.store(isize::try_from(new_width).unwrap(), Ordering::Relaxed);
+        let new_width = fish_wcstoi(&width_str.as_string()).unwrap_or(1).clamp(1, 2) as isize;
+        FISH_EMOJI_WIDTH.store(new_width, Ordering::Relaxed);
         FLOG!(
             term_support,
             "Overriding default fish_emoji_width w/",
@@ -167,39 +169,38 @@ fn guess_emoji_width(vars: &EnvStack) {
         return;
     }
 
-    let term = vars
+    let term_program = vars
         .get(L!("TERM_PROGRAM"))
         .map(|v| v.as_string())
         .unwrap_or_else(WString::new);
-    // The format and contents of $TERM_PROGRAM_VERSION depend on $TERM_PROGRAM. Under
-    // Apple_Terminal, this is an integral value in the hundreds corresponding to the
-    // CFBundleVersion of Terminal.app; under iTerm, this is the version number which can contain
-    // multiple periods (e.g 3.4.19). Currently we only care about Apple_Terminal but the C++ code
-    // used wcstod() to parse at least the major.minor value of cases like the latter.
-    //
-    // TODO: Move this inside the Apple_Terminal branch and use i32::FromStr (i.e. str::parse())
-    // instead.
-    let version = vars
-        .get(L!("TERM_PROGRAM_VERSION"))
-        .map(|v| v.as_string())
-        .and_then(|v| {
-            let mut consumed = 0;
-            crate::wutil::wcstod::wcstod(&v, '.', &mut consumed).ok()
-        })
-        .unwrap_or(0.0);
 
-    if term == "Apple_Terminal" && version as i32 >= 400 {
-        // Apple Terminal on High Sierra
-        FISH_EMOJI_WIDTH.store(2, Ordering::Relaxed);
-        FLOG!(term_support, "default emoji width: 2 for", term);
-    } else if term == "iTerm.app" {
+    #[allow(renamed_and_removed_lints)] // for old clippy
+    #[allow(clippy::blocks_in_if_conditions)] // for old clippy
+    if xtversion().unwrap_or(L!("")).starts_with(L!("iTerm2 ")) {
         // iTerm2 now defaults to Unicode 9 sizes for anything after macOS 10.12
         FISH_EMOJI_WIDTH.store(2, Ordering::Relaxed);
         FLOG!(term_support, "default emoji width 2 for iTerm2");
+    } else if term_program == "Apple_Terminal" && {
+        let version = vars
+            .get(L!("TERM_PROGRAM_VERSION"))
+            .map(|v| v.as_string())
+            .and_then(|v| {
+                let mut consumed = 0;
+                crate::wutil::wcstod::wcstod(&v, '.', &mut consumed).ok()
+            })
+            .unwrap_or(0.0);
+        version as i32 >= 400
+    } {
+        // Apple Terminal on High Sierra
+        FISH_EMOJI_WIDTH.store(2, Ordering::Relaxed);
+        FLOG!(term_support, "default emoji width: 2 for", term_program);
     } else {
         // Default to whatever the system's wcwidth gives for U+1F603, but only if it's at least
         // 1 and at most 2.
+        #[cfg(not(cygwin))]
         let width = crate::fallback::wcwidth('😃').clamp(1, 2);
+        #[cfg(cygwin)]
+        let width = 2_isize;
         FISH_EMOJI_WIDTH.store(width, Ordering::Relaxed);
         FLOG!(term_support, "default emoji width:", width);
     }
@@ -228,9 +229,8 @@ fn handle_change_ambiguous_width(vars: &EnvStack) {
         .and_then(|fish_ambiguous_width| fish_wcstoi(&fish_ambiguous_width).ok())
         .unwrap_or(1)
         // Clamp in case of negative values.
-        .max(0);
-    crate::fallback::FISH_AMBIGUOUS_WIDTH
-        .store(isize::try_from(new_width).unwrap(), Ordering::Relaxed);
+        .max(0) as isize;
+    crate::fallback::FISH_AMBIGUOUS_WIDTH.store(new_width, Ordering::Relaxed);
 }
 
 fn handle_term_size_change(vars: &EnvStack) {
@@ -282,6 +282,10 @@ fn handle_autosuggestion_change(vars: &EnvStack) {
     reader_set_autosuggestion_enabled(vars);
 }
 
+fn handle_transient_prompt_change(vars: &EnvStack) {
+    reader_set_transient_prompt(vars);
+}
+
 fn handle_function_path_change(_: &EnvStack) {
     function::invalidate_path();
 }
@@ -300,9 +304,11 @@ fn handle_locale_change(vars: &EnvStack) {
     guess_emoji_width(vars);
 }
 
-fn handle_curses_change(vars: &EnvStack) {
+fn handle_term_change(vars: &EnvStack) {
     guess_emoji_width(vars);
-    init_curses(vars);
+    init_terminal(vars);
+    read_terminfo_database(vars);
+    reader_schedule_prompt_repaint();
 }
 
 fn handle_fish_use_posix_spawn_change(vars: &EnvStack) {
@@ -354,7 +360,7 @@ pub fn env_dispatch_init(vars: &EnvStack) {
     use once_cell::sync::Lazy;
 
     run_inits(vars);
-    // env_dispatch_var_change() purposely supresses change notifications until the dispatch table
+    // env_dispatch_var_change() purposely suppresses change notifications until the dispatch table
     // was initialized elsewhere (either explicitly as below or via deref of VAR_DISPATCH_TABLE).
     Lazy::force(&VAR_DISPATCH_TABLE);
 }
@@ -362,7 +368,7 @@ pub fn env_dispatch_init(vars: &EnvStack) {
 /// Runs the subset of dispatch functions that need to be called at startup.
 fn run_inits(vars: &EnvStack) {
     init_locale(vars);
-    init_curses(vars);
+    init_terminal(vars);
     guess_emoji_width(vars);
     update_wait_on_escape_ms(vars);
     update_wait_on_sequence_key_ms(vars);
@@ -376,43 +382,23 @@ fn update_fish_color_support(vars: &EnvStack) {
     // Detect or infer term256 support. If fish_term256 is set, we respect it. Otherwise, infer it
     // from $TERM or use terminfo.
 
-    let term = vars
-        .get(L!("TERM"))
-        .map(|v| v.as_string())
-        .unwrap_or_else(WString::new);
-    let max_colors = curses::term().and_then(|term| term.max_colors);
-    let mut supports_256color = false;
-    let mut supports_24bit = false;
+    let term = vars.get_unless_empty(L!("TERM"));
+    let term = term.as_ref().map_or(L!(""), |term| &term.as_list()[0]);
+    let is_xterm_16color = term == "xterm-16color";
 
-    if let Some(fish_term256) = vars.get(L!("fish_term256")).map(|v| v.as_string()) {
-        // $fish_term256
-        supports_256color = crate::wcstringutil::bool_from_string(&fish_term256);
+    let supports_256color = if let Some(fish_term256) = vars.get(L!("fish_term256")) {
+        let ok = crate::wcstringutil::bool_from_string(&fish_term256.as_string());
         FLOG!(
             term_support,
             "256-color support determined by $fish_term256:",
-            supports_256color
+            ok
         );
-    } else if term.find(L!("256color")).is_some() {
-        // TERM contains "256color": 256 colors explicitly supported.
-        supports_256color = true;
-        FLOG!(term_support, "256-color support enabled for TERM", term);
-    } else if term.find(L!("xterm")).is_some() {
-        // Assume that all "xterm" terminals can handle 256
-        supports_256color = true;
-        FLOG!(term_support, "256-color support enabled for TERM", term);
-    }
-    // See if terminfo happens to identify 256 colors
-    else if let Some(max_colors) = max_colors {
-        supports_256color = max_colors >= 256;
-        FLOG!(
-            term_support,
-            "256-color support:",
-            max_colors,
-            "per termcap/terminfo entry for",
-            term
-        );
-    }
+        ok
+    } else {
+        !is_xterm_16color
+    };
 
+    let supports_24bit;
     if let Some(fish_term24bit) = vars.get(L!("fish_term24bit")).map(|v| v.as_string()) {
         // $fish_term24bit
         supports_24bit = crate::wcstringutil::bool_from_string(&fish_term24bit);
@@ -425,31 +411,14 @@ fn update_fish_color_support(vars: &EnvStack) {
                 "disabled"
             }
         );
-    } else if vars.get(L!("STY")).is_some() || term.starts_with(L!("eterm")) {
-        // Screen and emacs' ansi-term swallow true-color sequences, so we ignore them unless
-        // force-enabled.
+    } else if vars.get(L!("STY")).is_some() {
+        // Screen requires "truecolor on" to enable true-color sequences, so we ignore them
+        // unless force-enabled.
         supports_24bit = false;
-        FLOG!(
-            term_support,
-            "True-color support: disabled for eterm/screen"
-        );
-    } else if max_colors.unwrap_or(0) > 32767 {
-        // $TERM wins, xterm-direct reports 32767 colors and we assume that's the minimum as xterm
-        // is weird when it comes to color.
-        supports_24bit = true;
-        FLOG!(
-            term_support,
-            "True-color support: enabled per termcap/terminfo for",
-            term,
-            "with",
-            max_colors.unwrap(),
-            "colors"
-        );
+        FLOG!(term_support, "True-color support: disabled for screen");
     } else if let Some(ct) = vars.get(L!("COLORTERM")).map(|v| v.as_string()) {
         // If someone sets $COLORTERM, that's the sort of color they want.
-        if ct == "truecolor" || ct == "24bit" {
-            supports_24bit = true;
-        }
+        supports_24bit = ct == "truecolor" || ct == "24bit";
         FLOG!(
             term_support,
             "True-color support",
@@ -461,83 +430,54 @@ fn update_fish_color_support(vars: &EnvStack) {
             "per $COLORTERM",
             ct
         );
-    } else if vars.get(L!("KONSOLE_VERSION")).is_some()
-        || vars.get(L!("KONSOLE_PROFILE_NAME")).is_some()
-    {
-        // All Konsole versions that use $KONSOLE_VERSION are new enough to support this, so no
-        // check is needed.
-        supports_24bit = true;
-        FLOG!(term_support, "True-color support: enabled for Konsole");
-    } else if let Some(it) = vars.get(L!("ITERM_SESSION_ID")).map(|v| v.as_string()) {
-        // Supporting versions of iTerm include a colon here.
-        // We assume that if this is iTerm it can't also be st, so having this check inside is okay.
-        if !it.contains(':') {
-            supports_24bit = true;
-            FLOG!(term_support, "True-color support: enabled for iTerm");
-        }
-    } else if term.starts_with("st-") {
-        supports_24bit = true;
-        FLOG!(term_support, "True-color support: enabling for st");
-    } else if let Some(vte) = vars.get(L!("VTE_VERSION")).map(|v| v.as_string()) {
-        if fish_wcstoi(&vte).unwrap_or(0) > 3600 {
-            supports_24bit = true;
-            FLOG!(
-                term_support,
-                "True-color support: enabled for VTE version",
-                vte
-            );
-        }
+    } else {
+        supports_24bit = !is_xterm_16color
+            && !vars
+                .get_unless_empty(L!("TERM_PROGRAM"))
+                .is_some_and(|term| term.as_list()[0] == "Apple_Terminal");
+        FLOG!(
+            term_support,
+            "True-color support",
+            if supports_24bit {
+                "enabled"
+            } else {
+                "disabled"
+            },
+        );
     }
 
     let mut color_support = ColorSupport::default();
     color_support.set(ColorSupport::TERM_256COLOR, supports_256color);
     color_support.set(ColorSupport::TERM_24BIT, supports_24bit);
-    crate::output::set_color_support(color_support);
+    crate::terminal::set_color_support(color_support);
 }
 
-/// Apply any platform- or environment-specific hacks to our curses [`Term`] instance.
-fn apply_term_hacks(vars: &EnvStack, term: &mut Term) {
-    if cfg!(target_os = "macos") {
-        // Hack in missing italics and dim capabilities omitted from macOS xterm-256color terminfo.
-        // Improves the user experience under Terminal.app and iTerm.
-        let term_prog = vars
-            .get(L!("TERM_PROGRAM"))
-            .map(|v| v.as_string())
-            .unwrap_or(WString::new());
-        if term_prog == "Apple_Terminal" || term_prog == "iTerm.app" {
-            if let Some(term_val) = vars.get(L!("TERM")).map(|v| v.as_string()) {
-                if term_val == "xterm-256color" {
-                    const SITM_ESC: &[u8] = b"\x1B[3m";
-                    const RITM_ESC: &[u8] = b"\x1B[23m";
-                    const DIM_ESC: &[u8] = b"\x1B[2m";
+// Initialize the terminal subsystem
+fn init_terminal(vars: &EnvStack) {
+    let term = vars.get(L!("TERM"));
+    let term = term
+        .as_ref()
+        .and_then(|v| v.as_list().get(0))
+        .map(|v| v.as_utfstr())
+        .unwrap_or(L!(""));
 
-                    if term.enter_italics_mode.is_none() {
-                        term.enter_italics_mode = Some(SITM_ESC.to_cstring());
-                    }
-                    if term.exit_italics_mode.is_none() {
-                        term.exit_italics_mode = Some(RITM_ESC.to_cstring());
-                    }
-                    if term.enter_dim_mode.is_none() {
-                        term.enter_dim_mode = Some(DIM_ESC.to_cstring());
-                    }
-                }
-            }
-        }
-    }
-}
+    IS_DUMB.store(term == "dumb");
+    ONLY_GRAYSCALE.store(term == "ansi-m" || term == "linux-m" || term == "xterm-mono");
 
-/// Apply any platform- or environment-specific hacks that don't involve a `Term` instance.
-fn apply_non_term_hacks(vars: &EnvStack) {
-    // Midnight Commander tries to extract the last line of the prompt, and does so in a way that is
-    // broken if you do '\r' after it like we normally do.
-    // See https://midnight-commander.org/ticket/4258.
     if vars.get(L!("MC_SID")).is_some() {
         screen_set_midnight_commander_hack();
     }
+
+    update_fish_color_support(vars);
 }
 
-// Initialize the curses subsystem
-fn init_curses(vars: &EnvStack) {
+pub fn read_terminfo_database(vars: &EnvStack) {
+    if !use_terminfo() {
+        return;
+    }
+
+    // The current process' environment needs to be modified because the terminfo crate will
+    // read these variables
     for var_name in CURSES_VARIABLES {
         if let Some(value) = vars
             .getf_unless_empty(var_name, EnvMode::EXPORT)
@@ -551,37 +491,10 @@ fn init_curses(vars: &EnvStack) {
         }
     }
 
-    if curses::setup(None, |term| apply_term_hacks(vars, term)).is_none() {
-        if is_interactive_session() {
-            let term = vars.get_unless_empty(L!("TERM")).map(|v| v.as_string());
-            if let Some(term) = term {
-                FLOG!(
-                    term_support,
-                    wgettext_fmt!("Could not set up terminal for $TERM '%ls'. Falling back to hardcoded xterm-256color values", term)
-                );
-            } else {
-                FLOG!(
-                    term_support,
-                    wgettext!("Could not set up terminal because $TERM is unset. Falling back to hardcoded xterm-256color values")
-                );
-            }
-        }
+    terminal::setup();
 
-        curses::setup_fallback_term();
-    }
-
-    // Configure hacks that apply regardless of whether we successfully init curses or not.
-    apply_non_term_hacks(vars);
-
-    // Store some global variables that reflect the term's capabilities
-    if let Some(term) = curses::term() {
-        TERM_HAS_XN.store(term.eat_newline_glitch, Ordering::Relaxed);
-    }
-
-    update_fish_color_support(vars);
     // Invalidate the cached escape sequences since they may no longer be valid.
     LAYOUT_CACHE_SHARED.lock().unwrap().clear();
-    CURSES_INITIALIZED.store(true, Ordering::Relaxed);
 }
 
 /// Initialize the locale subsystem
@@ -687,18 +600,8 @@ fn init_locale(vars: &EnvStack) {
         new_msg_locale.to_string_lossy()
     );
 
-    #[cfg(gettext)]
-    {
-        if old_msg_locale.as_c_str() != new_msg_locale {
-            // Make change known to GNU gettext.
-            extern "C" {
-                static mut _nl_msg_cat_cntr: libc::c_int;
-            }
-            unsafe {
-                _nl_msg_cat_cntr += 1;
-            }
-        }
-    }
+    #[cfg(feature = "localize-messages")]
+    crate::wutil::gettext::update_locale_from_env(vars);
 }
 
 pub fn use_posix_spawn() -> bool {
@@ -706,6 +609,7 @@ pub fn use_posix_spawn() -> bool {
 }
 
 /// Whether or not we are running on an OS where we allow ourselves to use `posix_spawn()`.
+#[allow(clippy::needless_bool)] // for old clippy
 const fn allow_use_posix_spawn() -> bool {
     // OpenBSD's posix_spawn returns status 127 instead of erroring with ENOEXEC when faced with a
     // shebang-less script. Disable posix_spawn on OpenBSD.
