@@ -28,11 +28,14 @@ pub enum FeatureFlag {
     /// Remove `test`'s one and zero arg mode (make `test -n` return false etc)
     test_require_arg,
 
-    /// Whether keyboard protocols (kitty's CSI x u, xterm's modifyOtherKeys) are used
-    keyboard_protocols,
-
     /// Whether to write OSC 133 prompt markers
     mark_prompt,
+
+    /// Do not look up $TERM in terminfo database.
+    ignore_terminfo,
+
+    /// Whether we are allowed to query the TTY for extra information.
+    query_term,
 }
 
 struct Features {
@@ -114,18 +117,26 @@ pub const METADATA: &[FeatureMetadata] = &[
         read_only: false,
     },
     FeatureMetadata {
-        flag: FeatureFlag::keyboard_protocols,
-        name: L!("keyboard-protocols"),
+        flag: FeatureFlag::mark_prompt,
+        name: L!("mark-prompt"),
         groups: L!("4.0"),
-        description: L!("Use keyboard protocols (kitty, xterm's modifyotherkeys"),
+        description: L!("write OSC 133 prompt markers to the terminal"),
         default_value: true,
         read_only: false,
     },
     FeatureMetadata {
-        flag: FeatureFlag::mark_prompt,
-        name: L!("mark-prompt"),
-        groups: L!("4.0"),
-        description: L!("Write OSC 133 prompt markers to the terminal"),
+        flag: FeatureFlag::ignore_terminfo,
+        name: L!("ignore-terminfo"),
+        groups: L!("4.1"),
+        description: L!("do not look up $TERM in terminfo database"),
+        default_value: true,
+        read_only: false,
+    },
+    FeatureMetadata {
+        flag: FeatureFlag::query_term,
+        name: L!("query-term"),
+        groups: L!("4.1"),
+        description: L!("query the TTY to enable extra functionality"),
         default_value: true,
         read_only: false,
     },
@@ -192,6 +203,7 @@ impl Features {
                 AtomicBool::new(METADATA[5].default_value),
                 AtomicBool::new(METADATA[6].default_value),
                 AtomicBool::new(METADATA[7].default_value),
+                AtomicBool::new(METADATA[8].default_value),
             ],
         }
     }
