@@ -3,7 +3,7 @@
 //! path-related issues.
 
 use crate::common::{wcs2osstring, wcs2zstring};
-use crate::env::{EnvMode, EnvStack, Environment};
+use crate::env::{EnvMode, EnvStack, Environment, FALLBACK_PATH};
 use crate::expand::{expand_tilde, HOME_DIRECTORY};
 use crate::flog::{FLOG, FLOGF};
 use crate::wchar::prelude::*;
@@ -196,15 +196,6 @@ pub fn path_get_path(cmd: &wstr, vars: &dyn Environment) -> Option<WString> {
     }
 }
 
-// PREFIX is defined at build time.
-pub static DEFAULT_PATH: Lazy<[WString; 3]> = Lazy::new(|| {
-    [
-        WString::from_str(env!("PREFIX")) + L!("/bin"),
-        L!("/usr/bin").to_owned(),
-        L!("/bin").to_owned(),
-    ]
-});
-
 /// Finds the path of an executable named `cmd`, by looking in $PATH taken from `vars`.
 /// On success, err will be 0 and the path is returned.
 /// On failure, we return the "best path" with err set appropriately.
@@ -225,7 +216,7 @@ pub fn path_try_get_path(cmd: &wstr, vars: &dyn Environment) -> GetPathResult {
     if let Some(path) = vars.get(L!("PATH")) {
         path_get_path_core(cmd, path.as_list())
     } else {
-        path_get_path_core(cmd, &*DEFAULT_PATH)
+        path_get_path_core(cmd, &FALLBACK_PATH)
     }
 }
 
@@ -433,28 +424,22 @@ pub fn path_as_implicit_cd(path: &wstr, wd: &wstr, vars: &dyn Environment) -> Op
 pub fn path_make_canonical(path: &mut WString) {
     let chars: &mut [char] = path.as_char_slice_mut();
 
-    // Ignore trailing slashes, unless it's the first character.
-    let mut len = chars.len();
-    while len > 1 && chars[len - 1] == '/' {
-        len -= 1;
-    }
-
     // Turn runs of slashes into a single slash.
-    let mut trailing = 0;
+    let mut written = 0;
     let mut prev_was_slash = false;
-    for leading in 0..len {
-        let c = chars[leading];
+    for read in 0..chars.len() {
+        let c = chars[read];
         let is_slash = c == '/';
-        if !prev_was_slash || !is_slash {
-            // This is either the first slash in a run, or not a slash at all.
-            chars[trailing] = c;
-            trailing += 1;
+        if prev_was_slash && is_slash {
+            continue;
         }
+        // This is either the first slash in a run, or not a slash at all.
+        chars[written] = c;
+        written += 1;
         prev_was_slash = is_slash;
     }
-    assert!(trailing <= len);
-    if trailing < len {
-        path.truncate(trailing);
+    if written > 1 {
+        path.truncate(written - usize::from(prev_was_slash));
     }
 }
 
@@ -604,7 +589,7 @@ fn make_base_directory(xdg_var: &wstr, non_xdg_homepath: &wstr) -> BaseDirectory
         use std::path::PathBuf;
 
         let mut build_dir = PathBuf::from(env!("FISH_BUILD_DIR"));
-        build_dir.push("fish_root");
+        build_dir.push("fish-test-home");
 
         let err = match std::fs::create_dir_all(&build_dir) {
             Ok(_) => 0,
@@ -671,9 +656,9 @@ fn create_dir_all_with_mode<P: AsRef<std::path::Path>>(path: P, mode: u32) -> st
 }
 
 /// Return whether the given path is on a remote filesystem.
-fn path_remoteness(path: &wstr) -> DirRemoteness {
+pub fn path_remoteness(path: &wstr) -> DirRemoteness {
     let narrow = wcs2zstring(path);
-    #[cfg(target_os = "linux")]
+    #[cfg(any(target_os = "linux", cygwin))]
     {
         let mut buf = MaybeUninit::uninit();
         if unsafe { libc::statfs(narrow.as_ptr(), buf.as_mut_ptr()) } < 0 {
@@ -707,10 +692,10 @@ fn path_remoteness(path: &wstr) -> DirRemoteness {
             }
         }
     }
-    #[cfg(not(target_os = "linux"))]
+    #[cfg(not(any(target_os = "linux", cygwin)))]
     {
         fn remoteness_via_statfs<StatFS, Flags>(
-            statfn: unsafe extern "C" fn(*const i8, *mut StatFS) -> libc::c_int,
+            statfn: unsafe extern "C" fn(*const libc::c_char, *mut StatFS) -> libc::c_int,
             flagsfn: fn(&StatFS) -> Flags,
             is_local_flag: u64,
             path: &std::ffi::CStr,
@@ -813,6 +798,10 @@ fn test_path() {
     path_make_canonical(&mut path);
     assert_eq!(&path, L!("/"));
 
+    path = L!("/home/fishuser/").to_owned();
+    path_make_canonical(&mut path);
+    assert_eq!(&path, L!("/home/fishuser"));
+
     assert!(!paths_are_equivalent(L!("/foo/bar/baz"), L!("foo/bar/baz")));
     assert!(paths_are_equivalent(
         L!("///foo///bar/baz"),
