@@ -1,12 +1,13 @@
+use std::mem::MaybeUninit;
 use std::num::NonZeroI32;
 
 use crate::common::exit_without_destructors;
 use crate::event::{enqueue_signal, is_signal_observed};
 use crate::nix::getpid;
-use crate::panic::AT_EXIT;
-use crate::reader::{reader_handle_sigint, reader_sighup};
+use crate::reader::{reader_handle_sigint, reader_sighup, safe_restore_term_mode};
 use crate::termsize::TermsizeContainer;
 use crate::topic_monitor::{topic_monitor_principal, Generation, GenerationsList, Topic};
+use crate::tty_handoff::{safe_deactivate_tty_protocols, safe_mark_tty_invalid};
 use crate::wchar::prelude::*;
 use crate::wutil::{fish_wcstoi, perror};
 use errno::{errno, set_errno};
@@ -83,15 +84,15 @@ extern "C" fn fish_signal_handler(
             // Exit unless the signal was trapped.
             if !observed {
                 reader_sighup();
+                safe_mark_tty_invalid();
             }
             topic_monitor_principal().post(Topic::sighupint);
         }
         libc::SIGTERM => {
-            // Handle sigterm. The only thing we do is restore the front process ID, then die.
+            // Handle sigterm. The only thing we do is restore the front process ID and disable protocols, then die.
             if !observed {
-                if let Some(at_exit) = AT_EXIT.get() {
-                    (at_exit)(true);
-                }
+                safe_restore_term_mode();
+                safe_deactivate_tty_protocols();
                 // Safety: signal() and raise() are async-signal-safe.
                 unsafe {
                     libc::signal(libc::SIGTERM, libc::SIG_DFL);
@@ -131,8 +132,9 @@ pub fn signal_reset_handlers() {
 
     for data in SIGNAL_TABLE.iter() {
         if data.signal == libc::SIGHUP {
-            let mut oact: libc::sigaction = unsafe { std::mem::zeroed() };
-            unsafe { libc::sigaction(libc::SIGHUP, std::ptr::null(), &mut oact) };
+            let mut oact = MaybeUninit::uninit();
+            unsafe { libc::sigaction(libc::SIGHUP, std::ptr::null(), oact.as_mut_ptr()) };
+            let oact = unsafe { oact.assume_init() };
             if oact.sa_sigaction == libc::SIG_IGN {
                 continue;
             }
@@ -277,30 +279,31 @@ pub fn signal_handle(sig: Signal) {
 }
 
 pub static signals_to_default: Lazy<libc::sigset_t> = Lazy::new(|| {
-    let mut set: libc::sigset_t = unsafe { std::mem::zeroed() };
-    unsafe { libc::sigemptyset(&mut set) };
+    let mut set = MaybeUninit::uninit();
+    unsafe { libc::sigemptyset(set.as_mut_ptr()) };
     for data in SIGNAL_TABLE.iter() {
         // If SIGHUP is being ignored (e.g., because were were run via `nohup`) don't reset it.
         // We don't special case other signals because if they're being ignored that shouldn't
         // affect processes we spawn. They should get the default behavior for those signals.
         if data.signal == libc::SIGHUP {
-            let mut act: libc::sigaction = unsafe { std::mem::zeroed() };
-            unsafe { libc::sigaction(data.signal.code(), std::ptr::null(), &mut act) };
+            let mut act = MaybeUninit::uninit();
+            unsafe { libc::sigaction(data.signal.code(), std::ptr::null(), act.as_mut_ptr()) };
+            let act = unsafe { act.assume_init() };
             if act.sa_sigaction == libc::SIG_IGN {
                 continue;
             }
         }
-        unsafe { libc::sigaddset(&mut set, data.signal.code()) };
+        unsafe { libc::sigaddset(set.as_mut_ptr(), data.signal.code()) };
     }
-    return set;
+    unsafe { set.assume_init() }
 });
 
 /// Ensure we did not inherit any blocked signals. See issue #3964.
 pub fn signal_unblock_all() {
     unsafe {
-        let mut iset: libc::sigset_t = std::mem::zeroed();
-        libc::sigemptyset(&mut iset);
-        libc::sigprocmask(libc::SIG_SETMASK, &iset, std::ptr::null_mut());
+        let mut iset = MaybeUninit::uninit();
+        libc::sigemptyset(iset.as_mut_ptr());
+        libc::sigprocmask(libc::SIG_SETMASK, iset.as_ptr(), std::ptr::null_mut());
     }
 }
 
@@ -348,11 +351,11 @@ impl SigChecker {
 struct LookupEntry {
     signal: Signal,
     name: &'static wstr,
-    desc: &'static wstr, // Note: this needs to be translated via gettext before presenting it to the user.
+    desc: LocalizableString, // Note: this needs to be localized via gettext before presenting it to the user.
 }
 
 impl LookupEntry {
-    const fn new(signal: i32, name: &'static wstr, desc: &'static wstr) -> Self {
+    const fn new(signal: i32, name: &'static wstr, desc: LocalizableString) -> Self {
         Self {
             signal: Signal::new(signal),
             name,
@@ -361,66 +364,96 @@ impl LookupEntry {
     }
 }
 
+macro_rules! signal_entry {
+    ($name:ident, $desc:literal) => {
+        LookupEntry::new(
+            libc::$name,
+            L!(stringify!($name)),
+            localizable_string!($desc),
+        )
+    };
+    ($name:ident, $desc:expr) => {
+        LookupEntry::new(libc::$name, L!(stringify!($name)), $desc)
+    };
+}
+
 // Lookup table used to convert between signal names and signal ids, etc.
 #[rustfmt::skip]
 const SIGNAL_TABLE : &[LookupEntry] = &[
-    LookupEntry::new(libc::SIGHUP,    L!("SIGHUP"), L!("Terminal hung up")),
-    LookupEntry::new(libc::SIGINT,    L!("SIGINT"), L!("Quit request from job control (^C)")),
-    LookupEntry::new(libc::SIGQUIT,   L!("SIGQUIT"), L!("Quit request from job control with core dump (^\\)")),
-    LookupEntry::new(libc::SIGILL,    L!("SIGILL"), L!("Illegal instruction")),
-    LookupEntry::new(libc::SIGTRAP,   L!("SIGTRAP"), L!("Trace or breakpoint trap")),
-    LookupEntry::new(libc::SIGABRT,   L!("SIGABRT"), L!("Abort")),
-    LookupEntry::new(libc::SIGBUS,    L!("SIGBUS"), L!("Misaligned address error")),
-    LookupEntry::new(libc::SIGFPE,    L!("SIGFPE"), L!("Floating point exception")),
-    LookupEntry::new(libc::SIGKILL,   L!("SIGKILL"), L!("Forced quit")),
-    LookupEntry::new(libc::SIGUSR1,   L!("SIGUSR1"), L!("User defined signal 1")),
-    LookupEntry::new(libc::SIGUSR2,   L!("SIGUSR2"), L!("User defined signal 2")),
-    LookupEntry::new(libc::SIGSEGV,   L!("SIGSEGV"), L!("Address boundary error")),
-    LookupEntry::new(libc::SIGPIPE,   L!("SIGPIPE"), L!("Broken pipe")),
-    LookupEntry::new(libc::SIGALRM,   L!("SIGALRM"), L!("Timer expired")),
-    LookupEntry::new(libc::SIGTERM,   L!("SIGTERM"), L!("Polite quit request")),
-    LookupEntry::new(libc::SIGCHLD,   L!("SIGCHLD"), L!("Child process status changed")),
-    LookupEntry::new(libc::SIGCONT,   L!("SIGCONT"), L!("Continue previously stopped process")),
-    LookupEntry::new(libc::SIGSTOP,   L!("SIGSTOP"), L!("Forced stop")),
-    LookupEntry::new(libc::SIGTSTP,   L!("SIGTSTP"), L!("Stop request from job control (^Z)")),
-    LookupEntry::new(libc::SIGTTIN,   L!("SIGTTIN"), L!("Stop from terminal input")),
-    LookupEntry::new(libc::SIGTTOU,   L!("SIGTTOU"), L!("Stop from terminal output")),
-    LookupEntry::new(libc::SIGURG,    L!("SIGURG"), L!("Urgent socket condition")),
-    LookupEntry::new(libc::SIGXCPU,   L!("SIGXCPU"), L!("CPU time limit exceeded")),
-    LookupEntry::new(libc::SIGXFSZ,   L!("SIGXFSZ"), L!("File size limit exceeded")),
-    LookupEntry::new(libc::SIGVTALRM, L!("SIGVTALRM"), L!("Virtual timefr expired")),
-    LookupEntry::new(libc::SIGPROF,   L!("SIGPROF"), L!("Profiling timer expired")),
-    LookupEntry::new(libc::SIGWINCH,  L!("SIGWINCH"), L!("Window size change")),
-    LookupEntry::new(libc::SIGIO,     L!("SIGIO"), L!("I/O on asynchronous file descriptor is possible")),
-    LookupEntry::new(libc::SIGSYS,    L!("SIGSYS"), L!("Bad system call")),
-    LookupEntry::new(libc::SIGIOT,    L!("SIGIOT"), L!("Abort (Alias for SIGABRT)")),
+    signal_entry!(SIGHUP, "Terminal hung up"),
+    signal_entry!(SIGINT, "Quit request from job control (^C)"),
+    signal_entry!(SIGQUIT, "Quit request from job control with core dump (^\\)"),
+    signal_entry!(SIGILL, "Illegal instruction"),
+    signal_entry!(SIGTRAP, "Trace or breakpoint trap"),
+    signal_entry!(SIGABRT, "Abort"),
+    signal_entry!(SIGBUS, "Misaligned address error"),
+    signal_entry!(SIGFPE, "Floating point exception"),
+    signal_entry!(SIGKILL, "Forced quit"),
+    signal_entry!(SIGUSR1, "User defined signal 1"),
+    signal_entry!(SIGUSR2, "User defined signal 2"),
+    signal_entry!(SIGSEGV, "Address boundary error"),
+    signal_entry!(SIGPIPE, "Broken pipe"),
+    signal_entry!(SIGALRM, "Timer expired"),
+    signal_entry!(SIGTERM, "Polite quit request"),
+    signal_entry!(SIGCHLD, "Child process status changed"),
+    signal_entry!(SIGCONT, "Continue previously stopped process"),
+    signal_entry!(SIGSTOP, "Forced stop"),
+    signal_entry!(SIGTSTP, "Stop request from job control (^Z)"),
+    signal_entry!(SIGTTIN, "Stop from terminal input"),
+    signal_entry!(SIGTTOU, "Stop from terminal output"),
+    signal_entry!(SIGURG, "Urgent socket condition"),
+    signal_entry!(SIGXCPU, "CPU time limit exceeded"),
+    signal_entry!(SIGXFSZ, "File size limit exceeded"),
+    signal_entry!(SIGVTALRM, "Virtual timefr expired"),
+    signal_entry!(SIGPROF, "Profiling timer expired"),
+    signal_entry!(SIGWINCH, "Window size change"),
+    signal_entry!(SIGIO, "I/O on asynchronous file descriptor is possible"),
+    signal_entry!(SIGSYS, "Bad system call"),
+    signal_entry!(SIGIOT, "Abort (Alias for SIGABRT)"),
 
-    #[cfg(any(bsd, target_os = "macos"))]
-    LookupEntry::new(libc::SIGEMT,    L!("SIGEMT"), L!("Unused signal")),
+    #[cfg(any(apple, bsd))]
+    signal_entry!(SIGEMT, SIGEMT_DESC),
 
-    #[cfg(any(bsd, target_os = "macos"))]
-    LookupEntry::new(libc::SIGINFO,   L!("SIGINFO"), L!("Information request")),
+    #[cfg(any(apple, bsd))]
+    signal_entry!(SIGINFO, SIGINFO_DESC),
 
     #[cfg(target_os = "linux")]
-    LookupEntry::new(libc::SIGSTKFLT, L!("SISTKFLT"), L!("Stack fault")),
+    signal_entry!(SIGSTKFLT, SIGSTKFLT_DESC),
 
     #[cfg(target_os = "linux")]
-    LookupEntry::new(libc::SIGIOT,   L!("SIGIOT"), L!("Abort (Alias for SIGABRT)")),
+    signal_entry!(SIGIOT, SIGIOT_DESC),
 
     #[cfg(target_os = "linux")]
-    LookupEntry::new(libc::SIGPWR,    L!("SIGPWR"), L!("Power failure")),
+    signal_entry!(SIGPWR, SIGPWR_DESC),
 
     // TODO: determine whether SIGWIND is defined on any platform.
-    //LookupEntry::new(libc::SIGWIND,   L!("SIGWIND"), L!("Window size change")),
+    //signal_entry!(SIGWIND, "Window size change"),
 ];
 
+localizable_consts!(
+    #[allow(dead_code)]
+    SIGEMT_DESC "Unused signal"
+
+    #[allow(dead_code)]
+    SIGINFO_DESC "Information request"
+
+    #[allow(dead_code)]
+    SIGSTKFLT_DESC "Stack fault"
+
+    #[allow(dead_code)]
+    SIGIOT_DESC "Abort (Alias for SIGABRT)"
+
+    #[allow(dead_code)]
+    SIGPWR_DESC "Power failure"
+);
+
 // Return true if two strings are equal, ignoring ASCII case.
 fn equals_ascii_icase(left: &wstr, right: &wstr) -> bool {
     if left.len() != right.len() {
         return false;
     }
     for (lc, rc) in left.chars().zip(right.chars()) {
-        if lc.to_ascii_lowercase() != rc.to_ascii_lowercase() {
+        if !lc.eq_ignore_ascii_case(&rc) {
             return false;
         }
     }
@@ -470,7 +503,7 @@ impl Signal {
     /// Previously signal_get_desc().
     pub fn desc(&self) -> &'static wstr {
         match self.get_lookup_entry() {
-            Some(entry) => wgettext_str(entry.desc),
+            Some(entry) => entry.desc.localize(),
             None => wgettext!("Unknown"),
         }
     }
