@@ -9,6 +9,7 @@ use crate::tests::string_escape::ESCAPE_TEST_CHAR;
 use crate::util::get_rng;
 use crate::wchar::prelude::*;
 use crate::wcstringutil::{string_prefixes_string, string_prefixes_string_case_insensitive};
+use fish_build_helper::workspace_root;
 use rand::rngs::SmallRng;
 use rand::Rng;
 use std::collections::VecDeque;
@@ -272,8 +273,6 @@ fn test_history_races() {
     // Ensure history is clear.
     History::new(L!("race_test")).clear();
 
-    // history::CHAOS_MODE.store(true);
-
     let mut children = Vec::with_capacity(RACE_COUNT);
     for i in 0..RACE_COUNT {
         children.push(std::thread::spawn(move || {
@@ -295,7 +294,6 @@ fn test_history_races() {
 
     // Ensure that we got sane, sorted results.
     let hist = History::new(L!("race_test"));
-    history::CHAOS_MODE.store(false);
 
     // History is enumerated from most recent to least
     // Every item should be the last item in some array
@@ -312,18 +310,22 @@ fn test_history_races() {
                 continue;
             };
 
-            // Remove everything from this item on
-            let removed = list.splice(position.., []);
+            // Remove the item we found.
+            list.remove(position);
+
+            // We expected this item to be the last. Items after this item
+            // in this array were therefore not found in history.
+            let removed = list.drain(position..);
             for line in removed.into_iter().rev() {
-                println!("Item dropped from history: {line}");
+                printf!("Item dropped from history: %ls\n", line);
             }
 
             found = true;
             break;
         }
         if !found {
-            println!(
-                "Line '{}' found in history, but not found in some array",
+            printf!(
+                "Line '%ls' found in history, but not found in some array\n",
                 item.str()
             );
             for list in &expected_lines {
@@ -557,15 +559,15 @@ fn test_history_path_detection() {
 
     // Expected sets of paths.
     let expected_paths = [
-        vec![],                                   // cmd0
-        vec![filename.to_owned()],                // cmd1
-        vec![tmpdir + L!("/") + filename],        // cmd2
-        vec![L!("$HOME/").to_owned() + filename], // cmd3
-        vec![],                                   // cmd4
-        vec![L!("~/").to_owned() + filename],     // cmd5
-        vec![],                                   // cmd6
-        vec![],                                   // cmd7 - we do not expand globs
-        vec![],                                   // cmd8
+        vec![],                                    // cmd0
+        vec![filename.to_owned()],                 // cmd1
+        vec![tmpdir.clone() + L!("/") + filename], // cmd2
+        vec![L!("$HOME/").to_owned() + filename],  // cmd3
+        vec![],                                    // cmd4
+        vec![L!("~/").to_owned() + filename],      // cmd5
+        vec![],                                    // cmd6
+        vec![],                                    // cmd7 - we do not expand globs
+        vec![],                                    // cmd8
     ];
 
     let maxlap = 128;
@@ -586,14 +588,15 @@ fn test_history_path_detection() {
         std::thread::sleep(std::time::Duration::from_millis(2));
     }
     history.clear();
+    let _ = std::fs::remove_dir_all(wcs2osstring(&tmpdir));
 }
 
 fn install_sample_history(name: &wstr) {
     let path = path_get_data().expect("Failed to get data directory");
     std::fs::copy(
-        env!("CARGO_MANIFEST_DIR").to_owned()
-            + "/tests/"
-            + std::str::from_utf8(&wcs2string(name)).unwrap(),
+        workspace_root()
+            .join("tests")
+            .join(std::str::from_utf8(&wcs2string(name)).unwrap()),
         wcs2osstring(&(path + L!("/") + name + L!("_history"))),
     )
     .unwrap();
@@ -631,9 +634,7 @@ fn test_history_formats() {
         "echo foo".into(),
     ];
     let test_history_imported_from_bash = History::with_name(L!("bash_import"));
-    let file =
-        std::fs::File::open(env!("CARGO_MANIFEST_DIR").to_owned() + "/tests/history_sample_bash")
-            .unwrap();
+    let file = std::fs::File::open(workspace_root().join("tests/history_sample_bash")).unwrap();
     test_history_imported_from_bash.populate_from_bash(BufReader::new(file));
     assert_eq!(test_history_imported_from_bash.get_history(), expected);
     test_history_imported_from_bash.clear();
