@@ -168,12 +168,16 @@ sendline(r"""true; and builtin echo foo """)
 expect_prompt("bar")
 sendline(r"""abbr fruit --command={git,hg,svn} banana""")
 expect_prompt()
-sendline(r"""function git; echo git $argv; end; function hg; echo hg $argv; end; function svn; echo svn $argv; end""")
+sendline(
+    r"""function git; echo git $argv; end; function hg; echo hg $argv; end; function svn; echo svn $argv; end"""
+)
 expect_prompt()
 sendline(r"""git fruit""")
 expect_prompt("git banana")
 sendline(r"""abbr""")
-expect_prompt("abbr -a --position anywhere --command git --command hg --command svn -- fruit banana")
+expect_prompt(
+    "abbr -a --position anywhere --command git --command hg --command svn -- fruit banana"
+)
 sendline(r"""function banana; echo I am a banana; end""")
 expect_prompt()
 sendline(r"""abbr fruit --command={git,hg,svn,} banana""")
@@ -181,3 +185,10 @@ expect_prompt()
 sendline(r"""fruit foo""")
 expect_prompt("I am a banana")
 
+# FIXME: This fails regularly on CI, output like
+# \r\n\x1b]133;C;cmdline_url=echo%20bar\x07bar\r\n\x1b]133;D;0\x07\x1b[?25h⏎
+# {{spaces}}\r⏎ \r\rprompt 39>\x1b[?2004h
+# (don't add the literal string "bar" here or the expect_prompt will match it - so we add some no-op quotes)
+# sendline(r"""function replace; commandline -r ""; echo echo b''ar; end; abbr foo --function replace""")
+# sendline("foo")
+# expect_prompt("bar")
